import codecs
import commands

class Patcher(object):
    def patch (self, filename, begin, end, new_comment):
        file_encoding = commands.getoutput('file -b --mime-encoding %s' %
                filename)
        with codecs.open(filename, 'r', file_encoding) as f:
            lines = f.readlines()

        res = lines[0:begin] + [new_comment + '\n'] + lines[end:]
        res = ''.join(res)
        with open (filename, 'w') as f:
            f.write(res)



import argparse
import json
import glob
import os
from IPython.terminal.embed import InteractiveShellEmbed
from traitlets.config.loader import Config
from distutils.spawn import find_executable

QUICKSTART_HELP = \
"""
| Answer the prompts by returning a value.
|
| Example:
|
| What is your favorite folder ?
|
| >>> import os
| >>> res = os.path.abspath('.')
| >>> print res
| >>> res
|
| The answer will be the value of res, you will be prompted for confirmation.
|
| To skip a question, return None:
|
| >>> None
"""

class Skip(Exception):
    pass

class QuickStartShell(InteractiveShellEmbed):
    def __init__(self):
        cfg = Config()
        prompt_config = cfg.PromptManager
        prompt_config.out_template = 'Answer: '
        InteractiveShellEmbed.__init__(self, banner1='', config=cfg)

    def run_cell(self, raw_cell, **kwargs):
        if raw_cell.strip() == 'None':
            self.skip = True
            self.exit_now = True
            return None

        res = InteractiveShellEmbed.run_cell(self, raw_cell, **kwargs)

        if res.result:
            self.result = res.result
            self.exit_now = True

        return res

    def raw_input(self, prompt):
        res = None
        while res is None:
            try:
                res = InteractiveShellEmbed.raw_input(self, prompt)
            except KeyboardInterrupt:
                print "\nUse ctrl + D to quit"

        return res

    def ask_confirmation(self, prompt='Confirm [y,n]? '):
        res = None
        while res is None:
            user_res = self.raw_input(prompt)
            if user_res == 'y':
                res = True
            elif user_res == 'n':
                res = False
        return res

    def wait_for_continue(self, prompt='Press Enter to continue '):
        self.raw_input(prompt)
        return True

    def ask(self, question, confirm=False):
        self.skip = False
        self.result = None

        confirmed = False

        while not confirmed:
            self(header=question)

            if self.skip:
                raise Skip

            if self.result is None:
                raise EOFError

            confirmed = not confirm or self.ask_confirmation()

        return self.result


class QuickStartArgument(object):
    def __init__(self, wizard, argument, extra_prompt,
            validate_function, finalize_function):
        self.argument = argument
        self.wizard = wizard
        self.extra_prompt = extra_prompt
        self.validate_function = validate_function
        self.finalize_function = finalize_function

    def do_quick_start(self):
        return self.wizard.prompt_key(self.argument.dest,
                prompt='>>> %s ?' % self.argument.help,
                extra_prompt=self.extra_prompt,
                title=self.argument.help,
                validate_function=self.validate_function,
                finalize_function=self.finalize_function)

PROMPT_EXECUTABLE=\
"""
%s is needed for the setup to continue.

Press Enter once it is installed """

class QuickStartWizard(object):
    def __init__(self, group,
            parent=None,
            validate_function=None):
        self.group = group
        self._add_argument = group.add_argument
        group.add_argument = self._add_argument_override
        self._add_argument_group = group.add_argument_group
        group.add_argument_group = self._add_argument_group_override
        self._qs_objects = []
        self.parent = parent
        if parent is None:
            self.parent = self
            self.qsshell = QuickStartShell()
            self.config = {}
        else:
            self.qsshell = parent.qsshell
            self.config = parent.config

        self.validate_function = validate_function

    def group_prompt(self):
        self.before_prompt()

        if type(self.group) == argparse.ArgumentParser:
            title = self.group.prog
        else:
            title = self.group.title

        print "\nSetting up %s" % title
        if self.group.description:
            print "\nDescription: %s" % self.group.description

        print "Current configuration:\n"
        for qs_object in self._qs_objects:
            if type(qs_object) == QuickStartArgument:
                current_value = self.config.get(qs_object.argument.dest)
                print '%s : %s' % (qs_object.argument.dest, current_value)

        question = '\nWould you like to configure %s [y,n]? ' % title

        return self.ask_confirmation(question)

    def main_prompt(self):
        return self.group_prompt()

    def before_prompt(self):
        pass

    def quick_start(self):
        try:
            if not self.do_quick_start():
                return False
        except EOFError:
            print "Aborting quick start"
            return False

        return True

    def before_quick_start(self, obj):
        return

    def do_quick_start(self):
        if self == self.parent:
            prompt = self.main_prompt
        else:
            prompt = self.group_prompt

        if not prompt():
            return False

        for obj in self._qs_objects:
            try:
                self.before_quick_start(obj)
                obj.do_quick_start()
            except Skip:
                pass

        if self.validate_function:
            if not self.validate_function(self):
                return self.do_quick_start()

        return True

    def propose_choice(self, choices, skippable=True, extra_prompt=None):
        self.before_prompt()

        if extra_prompt:
            print extra_prompt

        skip_choice = -1

        if skippable:
            skip_choice = len(choices)
            choices.append("Skip")

        prompt = "Make your choice ["

        for i, choice in enumerate(choices):
            print '%s) %s' % (str(i), choice)
            if i != 0:
                prompt += ','
            prompt += str(i)

        print ""

        prompt += ']? '

        valid_choice = False

        while not valid_choice:
            res = self.qsshell.raw_input(prompt)
            try:
                valid_choice = int(res) in range(len(choices))
            except:
                valid_choice = False

            if not valid_choice:
                print "Invalid choice %s" % res

        if int(res) == skip_choice:
            raise Skip

        return int(res)

    def ask_confirmation(self, prompt='Confirm [y,n]? '):
        res = None
        while res is None:
            user_res = self.qsshell.raw_input(prompt)
            if user_res == 'y':
                res = True
            elif user_res == 'n':
                res = False
        return res

    def prompt_key(self, key, prompt=None, extra_prompt=None, title=None,
            store=True, validate_function=None, finalize_function=None):
        self.before_prompt()

        if title is None:
            title = key

        if key in self.config:
            print 'Current value for %s : %s' % (title,
                    self.config[key])
            if not self.qsshell.ask_confirmation('Update [y,n]? '):
                return self.config[key]

        if extra_prompt:
            print extra_prompt

        if not prompt:
            prompt = 'New value for %s? ' % title

        validated = False

        while not validated:
            res = self.qsshell.ask(prompt)
            if validate_function is None:
                validated = True
            else:
                validated = validate_function(self, res)

        if finalize_function:
            res = finalize_function(self, res)

        if store:
            self.config[key] = res

        return res

    def wait_for_continue(self, prompt='Press Enter to continue '):
        self.before_prompt()
        res = self.qsshell.raw_input(prompt)
        if res.lower() == "you're a wizard harry":
            print "Indeed"
        return True

    def prompt_executable(self, executable, prompt=None):
        if prompt is None:
            prompt = PROMPT_EXECUTABLE % executable

        while not find_executable(executable):
            self.wait_for_continue(prompt)

    @staticmethod
    def validate_folder(wizard, path):
        res = True
        if not os.path.exists(path):
            os.mkdir(path)
            res = True
        elif not os.path.isdir(path):
            print "Path %s is not a folder" % path
            res = False

        return res

    @staticmethod
    def validate_list(wizard, thing):
        res = type(thing) == list
        if not res:
            print "%s is not a list" % thing
        return res

    @staticmethod
    def check_path_is_file(wizard, path):
        res = True

        if not os.path.exists(path):
            print "Path %s does not exist" % path
            res = False
        elif not os.path.isfile(path):
            print "Path %s is not a file" % path
            res = False

        return res

    @staticmethod
    def validate_globs_list(wizard, thing):
        if not wizard.validate_list(wizard, thing):
            return False

        resolved = []
        for item in thing:
            resolved.extend (glob.glob(item))

        print "The expression(s) currently resolve(s) to:"
        print resolved

        return wizard.ask_confirmation()

    def _add_argument_group_override(self, parser, *args, **kwargs):
        validate_function = kwargs.pop('validate_function', None)
        wizard_class = kwargs.pop('wizard_class', type(self.parent))
        res = self._add_argument_group(parser, *args, **kwargs)
        wizard = wizard_class(res,
                parent=self.parent,
                validate_function=validate_function)
        self._qs_objects.append(wizard)
        return res

    def _add_argument_override(self, group, *args, **kwargs):
        validate_function = kwargs.pop('validate_function', None)
        finalize_function = kwargs.pop('finalize_function', None)
        extra_prompt = kwargs.pop('extra_prompt', None)
        no_prompt = kwargs.pop('no_prompt', False)
        arg = self._add_argument(group, *args, **kwargs)

        if not no_prompt:
            self._qs_objects.append(QuickStartArgument(self, arg,
                extra_prompt, validate_function, finalize_function))

        return arg



import os
import shutil

# FIXME: be less naive :)
class NaiveIndexFormatter(object):
    def __init__(self, symbols, directory='tmp_markdown_files',
            index_name='tmp-index.markdown'):
        pages = {}

        for name, symbol in symbols.iteritems():
            filename = symbol.filename
            if filename is None:
                continue
            page_symbols = pages.get(filename)
            if not page_symbols:
                page_symbols = []
                pages[filename] = page_symbols
            page_symbols.append (name)

        mddir = directory

        try:
            shutil.rmtree (mddir)
        except OSError as e:
            pass

        try:
            os.mkdir (mddir)
        except OSError:
            pass

        with open (os.path.join (mddir, index_name), 'w') as index:
            for page, symbols in pages.iteritems():
                base_name = os.path.basename (os.path.splitext(page)[0])
                filename = '%s.markdown' % base_name
                index.write ('#### [%s](%s)\n' % (base_name, filename))
                with open (os.path.join(mddir, filename), 'a') as f:
                    for symbol in symbols:
                        f.write('* [%s]()\n' % symbol)



import pygit2 as git
import os

class GitInterface(object):
    def __init__(self, repo_path=None):
        if repo_path:
            self.repo_path = os.path.abspath(repo_path)
            self.repo = git.Repository(self.repo_path)
            self.index = self.repo.index
            self.commiter = git.Signature('hotdoc', 'hotdoc@hotdoc.net')
        else:
            self.repo_path = None

    def set_repo_path(self, repo_path):
        self.repo_path = os.path.abspath(repo_path)
        self.repo = git.Repository(self.repo_path)
        self.index = self.repo.index
        self.commiter = git.Signature('hotdoc', 'hotdoc@hotdoc.net')

    def add (self, filename):
        filename = os.path.relpath(filename, self.repo_path)
        self.index.add(filename)
        self.index.write()

    def commit (self, author, author_mail, message):
        author = git.Signature(author, author_mail)
        tree = self.index.write_tree()
        oid = self.repo.create_commit('refs/heads/master',
                author, self.commiter, message, tree,
                [self.repo.head.get_object().hex])
        return oid



# -*- coding: utf-8 -*-

from sqlalchemy import (Column, String, Integer, Boolean,
        ForeignKey, PickleType)

from hotdoc.core.links import Link
from hotdoc.core.alchemy_integration import *
from hotdoc.core.comment_block import comment_from_tag


class Symbol (Base):
    __tablename__ = 'symbols'

    id = Column(Integer, primary_key=True)
    comment = Column(PickleType)
    unique_name = Column(String)
    display_name = Column(String)
    filename = Column(String)
    lineno = Column(Integer)
    location = Column(PickleType)
    _type_ = Column(String)
    extension_contents = Column(MutableDict.as_mutable(PickleType))
    extension_attributes = Column(MutableDict.as_mutable(PickleType))
    link = Column(Link.as_mutable(PickleType))
    skip = Column(Boolean)

    __mapper_args__ = {
            'polymorphic_identity': 'symbol',
            'polymorphic_on': _type_,
    }

    def __init__(self, **kwargs):
        self.extension_contents = {}
        self.extension_attributes = {}
        self.skip = False

        Base.__init__(self, **kwargs)

    # FIXME: this is a bit awkward to use.
    def add_extension_attribute (self, ext_name, key, value):
        attributes = self.extension_attributes.pop (ext_name, {})
        attributes[key] = value
        self.extension_attributes[ext_name] = attributes

    def get_extension_attribute (self, ext_name, key):
        attributes = self.extension_attributes.get (ext_name)
        if not attributes:
            return None
        return attributes.get (key)

    def get_children_symbols (self):
        return []

    def update_children_comments(self):
        for sym in self.get_children_symbols():
            if type(sym) == ParameterSymbol:
                sym.comment = self.comment.params.get(sym.argname)
            elif type(sym) == FieldSymbol:
                sym.comment = self.comment.params.get(sym.member_name)
            elif type(sym) == ReturnValueSymbol:
                tag = self.comment.tags.get('returns')
                sym.comment = comment_from_tag(tag)
            elif type(sym) == Symbol:
                sym.comment = self.comment.params.get(sym.display_name)

    def _make_name (self):
        return self.display_name

    def get_extra_links (self):
        return []

    def get_type_name (self):
        return ''

    def resolve_links(self, link_resolver):
        if self.link is None:
            self.link = Link(self.unique_name, self._make_name(),
                        self.unique_name)
        self.link = link_resolver.upsert_link(self.link, overwrite_ref=True)

class QualifiedSymbol (MutableObject):
    def __init__(self, type_tokens=[]):
        self.input_tokens = type_tokens
        self.comment = None
        self.extension_attributes = MutableDict()
        self.constructed()
        MutableObject.__init__(self)

    def add_extension_attribute (self, ext_name, key, value):
        attributes = self.extension_attributes.pop (ext_name, {})
        attributes[key] = value
        self.extension_attributes[ext_name] = attributes

    def get_extension_attribute (self, ext_name, key):
        attributes = self.extension_attributes.get (ext_name)
        if not attributes:
            return None
        return attributes.get (key)

    def get_children_symbols(self):
        return []

    def get_type_link (self):
        return self.type_link

    def resolve_links(self, link_resolver):
        self.type_link = None
        self.type_tokens = []

        for tok in self.input_tokens:
            if isinstance(tok, Link):
                self.type_link = link_resolver.get_named_link(tok.id_)
                if not self.type_link:
                    self.type_link = link_resolver.upsert_link(tok)
                self.type_tokens.append (self.type_link)
            else:
                self.type_tokens.append (tok)

    def constructed(self):
        self.extension_contents = {}

    def __setstate__(self, state):
        MutableObject.__setstate__(self, state)
        self.constructed()

class ReturnValueSymbol (QualifiedSymbol):
    def __init__(self, comment=None, **kwargs):
        QualifiedSymbol.__init__(self, **kwargs)
        self.comment = comment

class ParameterSymbol (QualifiedSymbol):
    def __init__(self, argname='', comment=None, **kwargs):
        QualifiedSymbol.__init__(self, **kwargs)
        # FIXME: gir specific
        self.array_nesting = 0
        self.argname = argname
        self.comment = comment

class FieldSymbol (QualifiedSymbol):
    def __init__(self, member_name='', is_function_pointer=False,
            comment=None, **kwargs):
        QualifiedSymbol.__init__(self, **kwargs)
        self.member_name = member_name
        self.is_function_pointer = is_function_pointer
        self.comment = comment

    def _make_name (self):
        return self.member_name

    def get_type_name (self):
        return "Attribute"

class FunctionSymbol (Symbol):
    __tablename__ = 'functions'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    parameters = Column(MutableList.as_mutable(PickleType))
    return_value = Column(ReturnValueSymbol.as_mutable(PickleType))
    is_method = Column(Boolean)
    throws = Column(Boolean)
    __mapper_args__ = {
            'polymorphic_identity': 'functions',
    }

    def __init__(self, **kwargs):
        self.parameters = []
        self.throws = False
        self.is_method = False
        Symbol.__init__(self, **kwargs)

    def get_children_symbols(self):
        return self.parameters + [self.return_value]

    def get_type_name (self):
        if self.is_method:
            return 'Method'
        return 'Function'

class SignalSymbol (FunctionSymbol):
    __tablename__ = 'signals'
    id = Column(Integer, ForeignKey('functions.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'signals',
    }
    flags = Column(PickleType)

    def __init__(self, **kwargs):
        # FIXME: flags are gobject-specific
        self.flags = []
        FunctionSymbol.__init__(self, **kwargs)

    def get_type_name (self):
        return "Signal"

class VFunctionSymbol (FunctionSymbol):
    __tablename__ = 'vfunctions'
    id = Column(Integer, ForeignKey('functions.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'vfunctions',
    }
    flags = Column(PickleType)

    def __init__(self, **kwargs):
        self.flags = []
        FunctionSymbol.__init__(self, **kwargs)

    def get_type_name (self):
        return "Virtual Method"

class PropertySymbol (Symbol):
    __tablename__ = 'properties'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'properties',
    }
    prop_type = Column(PickleType)

    def get_children_symbols(self):
        return [self.prop_type]

class CallbackSymbol (FunctionSymbol):
    __tablename__ = 'callbacks'
    id = Column(Integer, ForeignKey('functions.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'callbacks',
    }

    def get_type_name (self):
        return "Callback"

class EnumSymbol (Symbol):
    __tablename__ = 'enums'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'enums',
    }
    members = Column(PickleType)

    def __init__(self, **kwargs):
        self.members = {}
        Symbol.__init__(self, **kwargs)

    def get_children_symbols(self):
        return self.members

    def get_extra_links (self):
        return [m.link for m in self.members]

    def get_type_name (self):
        return "Enumeration"

class StructSymbol (Symbol):
    __tablename__ = 'structs'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'structs',
    }
    members = Column(PickleType)
    raw_text = Column(String)

    def __init__(self, **kwargs):
        self.members = {}
        Symbol.__init__(self, **kwargs)

    def get_children_symbols(self):
        return self.members

    def get_type_name (self):
        return "Structure"

# FIXME: and this is C-specific
class MacroSymbol (Symbol):
    __tablename__ = 'macros'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'macros',
    }
    original_text = Column(String)

class FunctionMacroSymbol (MacroSymbol):
    __tablename__ = 'function_macros'
    id = Column(Integer, ForeignKey('macros.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'function_macros',
    }

    parameters = Column(MutableList.as_mutable(PickleType))
    return_value = Column(PickleType)

    def __init__(self, **kwargs):
        self.parameters = []
        MacroSymbol.__init__(self, **kwargs)

    def get_children_symbols(self):
        return self.parameters + [self.return_value]

    def get_type_name (self):
        return "Function macro"

class ConstantSymbol (MacroSymbol):
    __tablename__ = 'constants'
    id = Column(Integer, ForeignKey('macros.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'constants',
    }

    def get_type_name (self):
        return "Constant"


class ExportedVariableSymbol (MacroSymbol):
    __tablename__ = 'exported_variables'
    id = Column(Integer, ForeignKey('macros.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'exported_variables',
    }
    type_qs = Column(PickleType)

    def get_type_name (self):
        return "Exported variable"

    def get_children_symbols(self):
        return [self.type_qs]

class AliasSymbol (Symbol):
    __tablename__ = 'aliases'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'aliases',
    }
    aliased_type = Column(PickleType)

    def get_type_name (self):
        return "Alias"

    def get_children_symbols(self):
        return [self.aliased_type]

class ClassSymbol (Symbol):
    __tablename__ = 'classes'
    id = Column(Integer, ForeignKey('symbols.id'), primary_key=True)
    __mapper_args__ = {
            'polymorphic_identity': 'classes',
    }
    # FIXME: multiple inheritance
    hierarchy = Column(PickleType)
    children = Column(PickleType)

    def __init__(self, **kwargs):
        self.hierarchy = []
        self.children = {}
        Symbol.__init__(self, **kwargs)

    def get_type_name (self):
        return "Class"

    def get_children_symbols(self):
        return self.hierarchy + list(self.children.values())



from bisect import bisect_left, bisect_right

class VersionList(list):
    def index(a, x):
        'Locate the leftmost value exactly equal to x'
        i = bisect_left(a, x)
        if i != len(a) and a[i] == x:
            return i
        raise ValueError

    def find_lt(a, x):
        'Find rightmost value less than x'
        i = bisect_left(a, x)
        if i:
            return a[i-1]
        raise ValueError

    def find_le(a, x):
        'Find rightmost value less than or equal to x'
        i = bisect_right(a, x)
        if i:
            return a[i-1]
        raise ValueError

    def find_gt(a, x):
        'Find leftmost value greater than x'
        i = bisect_right(a, x)
        if i != len(a):
            return a[i]
        raise ValueError

    def find_ge(a, x):
        'Find leftmost item greater than or equal to x'
        i = bisect_left(a, x)
        if i != len(a):
            return a[i]
        raise ValueError



import os
import os
import linecache

class TagValidator(object):
    def __init__(self, name):
        self.name = name

    def validate(self, value):
        raise NotImplementedError

class Comment(object):
    def __init__(self, name='', title='', params={}, filename='',
            lineno=-1, endlineno=-1, annotations={}, description='', short_description='',
            tags={}, raw_comment=''):
        self.name = name
        self.title = title
        self.params = params
        self.filename = os.path.abspath(filename)
        self.lineno = lineno
        self.endlineno = endlineno
        self.annotations = annotations
        self.description = description
        self.short_description = short_description
        self.tags = tags

        # FIXME : would be nicer to have the scanner do that ^^
        leading_ws = ''
        if lineno != -1:
            orig = linecache.getline(filename, lineno)
            leading_ws = (len(orig) - len (orig.lstrip(' '))) * ' '

        self.raw_comment = leading_ws + raw_comment

    def add_param_block (self, param_name, block):
        self.params[param_name] = block

    def set_return_block (self, block):
        self.tags['returns'] = block

    def set_description (self, description):
        self.description = description.strip();

class Annotation(object):
    def __init__(self, name, argument=None):
        self.name = name
        self.argument = argument

class Tag(object):
    def __init__(self, name, description, value=None, annotations={}):
        self.name = name
        self.description = description
        self.annotations = annotations

def comment_from_tag(tag):
    if not tag:
        return None
    comment = Comment (name=tag.name,
            description=tag.description,
            annotations=tag.annotations)
    return comment



import sys, re
from itertools import izip_longest
from hotdoc.core.comment_block import *

#http://stackoverflow.com/questions/434287/what-is-the-most-pythonic-way-to-iterate-over-a-list-in-chunks
def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return izip_longest(*args, fillvalue=fillvalue)

class GtkDocRawCommentParser (object):
    def __init__(self, doc_tool):
        """
        Lifted from
        http://stackoverflow.com/questions/5323703/regex-how-to-match-sequence-of-key-value-pairs-at-end-of-string
        """
        self.kv_regex = re.compile(r'''
                                   [\S]+=
                                   (?:
                                   \s*
                                   (?!\S+=)\S+
                                   )+
                                   ''', re.VERBOSE)
        self.doc_tool = doc_tool

        tag_validation_regex = r'((?:^|\n)[ \t]*('
        tag_validation_regex += 'returns|Returns|since|Since|deprecated|Deprecated|stability|Stability|Return value'
        for validator in doc_tool.tag_validators.values():
            tag_validation_regex += '|%s|%s' % (validator.name,
                    validator.name.lower())
        tag_validation_regex += '):)'

        self.tag_validation_regex = re.compile (tag_validation_regex)

    def parse_title (self, raw_title):
        # Section comments never contain annotations,
        # We also normalize them here to not contain any spaces.
        # FIXME: This code now only lives here for the purpose of gtk-doc conversion
        # Could remove that one day
        if "SECTION" in raw_title:
            return ''.join(raw_title.split(' ')), []

        split = raw_title.split (': ', 1)
        title = split[0].rstrip(':')
        annotations = []
        if len (split) > 1:
            annotations = self.parse_annotations (split[1])
        elif not raw_title.strip().endswith(':'):
            print "Missing colon in comment name :", raw_title
            return None, []
        return title, annotations

    def parse_key_value_annotation (self, name, string):
        arg = {}
        kvs = self.kv_regex.findall (string)
        kvs = dict([kv.split('=', 1) for kv in kvs])
        return Annotation (name=name, argument=kvs)

    def parse_annotation (self, string):
        split = string.split ()
        name = split[0].strip()
        if len (split) == 1:
            return Annotation (name=name)
        elif '=' in split[1]:
            return self.parse_key_value_annotation (name, split[1])
        else:
            return Annotation (name=name, argument=[split[1]])

    def parse_annotations (self, string):
        parsed_annotations = []
        par_level = 0
        current_annotation = ""
        for c in string:
            if c == '(':
                par_level += 1
            elif c == ')':
                par_level -= 1

            if par_level > 1:
                return []
            elif par_level == 1 and c not in '()':
                current_annotation += c
            elif par_level == 0:
                if c not in ' \t\n\r()':
                    return []
                if current_annotation:
                    ann = self.parse_annotation (current_annotation)
                    if ann:
                        parsed_annotations.append (ann)
                    current_annotation = ""
            elif par_level < 0:
                return []

        if par_level != 0:
            return []

        return parsed_annotations

    def extract_annotations (self, desc):
        split = desc.split (': ', 1)

        if len (split) == 1:
            return desc, []

        annotations = self.parse_annotations (split[0])
        if not annotations:
            return desc, []

        return split[1].strip(), annotations

    def parse_parameter (self, name, desc):
        name = name.strip()[1:-1].strip()
        desc = desc.strip()
        desc, annotations = self.extract_annotations (desc)
        annotations = {annotation.name: annotation for annotation in
                annotations}
        return Comment (name=name, annotations=annotations,
                description=desc)

    def parse_title_and_parameters (self, tp):
        tps = re.split (r'(\n[ \t]*@[\S]+[ \t]*:)', tp)
        title, annotations = self.parse_title (tps[0])
        parameters = []
        for name, desc in grouper (tps[1:], 2):
            parameters.append (self.parse_parameter (name, desc))
        return title, parameters, annotations

    def parse_since_tag (self, name, desc):
        return Tag (name=name, description=desc)

    def parse_deprecated_tag (self, name, desc):
        return Tag (name=name, description=desc)

    def parse_stability_tag (self, name, desc):
        return Tag (name=name, description=desc)

    def parse_returns_tag (self, name, desc):
        desc, annotations = self.extract_annotations (desc)
        annotations = {annotation.name: annotation for annotation in
                annotations}
        return Tag (name=name, annotations=annotations, description=desc)

    def parse_tag (self, name, desc):
        if name.lower() == "since":
            return self.parse_since_tag (name, desc)
        elif name.lower() == "returns":
            return self.parse_returns_tag (name, desc)
        elif name.lower() == "return value":
            return self.parse_returns_tag ("returns", desc)
        elif name.lower() == "stability":
            return self.parse_stability_tag ("stability", desc)
        elif name.lower() == "deprecated":
            return self.parse_deprecated_tag("deprecated", desc)
        else:
            validator = self.doc_tool.tag_validators.get(name)
            if not validator:
                print "FIXME no tag validator"
                return None
            if not validator.validate(desc):
                print "invalid value for tag %s : %s" % name, desc
                return None
            return Tag(name=name, description=desc)

    def parse_description_and_tags (self, dt):
        dts = self.tag_validation_regex.split(dt)
        tags = []

        desc = dts[0]
        if len (dts) == 1:
            return desc, tags

        for raw, name, tag_desc in grouper (dts[1:], 3):
            tag = self.parse_tag (name.strip(), tag_desc.strip())
            if tag:
                tags.append(tag)

        return desc, tags

    def strip_comment (self, comment):
        comment = re.sub ('^[\W]*\/[\*]+[\W]*', '', comment)
        comment = re.sub ('\*\/[\W]*$', '', comment)
        comment = re.sub ('\n[ \t]*\*', '\n', comment)
        return comment.strip()

    def validate_c_comment(self, comment):
        return re.match(r'(/\*\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/)$',
                comment) is not None

    def parse_comment (self, comment, filename, lineno, endlineno, stripped=False):
        if not stripped and not comment.lstrip().startswith ('/**'):
            return None

        if not self.validate_c_comment(comment.strip()):
            return None

        raw_comment = comment
        comment = unicode(comment.encode('utf8'))
        if not stripped:
            comment = self.strip_comment (comment)

        comment = unicode(comment.encode('utf8'))

        split = re.split (r'\n[\W]*\n', comment, maxsplit=1)
        block_name, parameters, annotations = self.parse_title_and_parameters (split[0])

        if not block_name:
            return None

        description = ""
        tags = []
        if len (split) > 1:
            description, tags = self.parse_description_and_tags (split[1])

        title = None
        short_description = None
        actual_parameters = {}
        for param in parameters:
            if param.name.lower() == 'short_description':
                short_description = param.description
            elif param.name.lower() == 'title':
                title = param.description
            else:
                actual_parameters[param.name] = param

        annotations = {annotation.name: annotation for annotation in
                annotations}
        tags = {tag.name.lower(): tag for tag in tags}

        block = Comment (name=block_name, filename=filename, lineno=lineno,
                endlineno=endlineno,
                annotations=annotations, params=actual_parameters,
                description = description, short_description=short_description,
                title=title, tags=tags, raw_comment=raw_comment)

        return block

if __name__ == "__main__":
    dp = GtkDocRawCommentParser()
    with open (sys.argv[1], 'r') as f:
        c = f.read()
        block = dp.parse_comment (c, sys.argv[1], 0)



import os, sys, argparse

reload(sys)  
sys.setdefaultencoding('utf8')

import cPickle as pickle
import glob
import json

from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker, mapper

from .naive_index import NaiveIndexFormatter
from .links import LinkResolver
from .symbols import *
from .base_extension import BaseExtension
from .alchemy_integration import Base
from .doc_tree import DocTree
from .comment_block import Tag, Comment

from ..utils.wizard import QuickStartWizard, QUICKSTART_HELP, Skip, QuickStartArgument
from ..utils.utils import all_subclasses
from ..utils.simple_signals import Signal
from ..utils.loggable import TerminalController
from ..utils.utils import get_all_extension_classes
from ..formatters.html.html_formatter import HtmlFormatter

try:
    from ..utils.git_interface import GitInterface
    HAVE_GIT_INTERFACE = True
except ImportError:
    HAVE_GIT_INTERFACE = False

from hotdoc.core.gi_raw_parser import GtkDocRawCommentParser

class ConfigError(Exception):
    pass

class ChangeTracker(object):
    def __init__(self):
        self.exts_mtimes = {}

    def update_extension_sources_mtimes(self, extension):
        ext_mtimes = {}
        source_files = extension.get_source_files()
        for source_file in source_files:
            mtime = os.path.getmtime(source_file)
            ext_mtimes[source_file] = mtime

        self.exts_mtimes[extension.EXTENSION_NAME] = ext_mtimes

    def mark_extension_stale_sources (self, extension):
        stale = []
        source_files = extension.get_source_files()

        if extension.EXTENSION_NAME in self.exts_mtimes:
            prev_mtimes = self.exts_mtimes[extension.EXTENSION_NAME]
        else:
            prev_mtimes = {}

        for source_file in source_files:
            if not source_file in prev_mtimes:
                stale.append(source_file)
            else:
                prev_mtime = prev_mtimes.get(source_file)
                mtime = os.path.getmtime(source_file)
                if prev_mtime != mtime:
                    stale.append(source_file)

        extension.set_stale_source_files(stale)

HOTDOC_ASCII =\
"""/**    __  __   ____     ______   ____     ____     ______
 *    / / / /  / __ \   /_  __/  / __ \   / __ \   / ____/
 *   / /_/ /  / / / /    / /    / / / /  / / / /  / /     
 *  / __  /  / /_/ /    / /    / /_/ /  / /_/ /  / /___   
 * /_/ /_/   \____/    /_/    /_____/   \____/   \____/   
 *
 * The Tastiest Documentation Tool.
 */                                                      
"""

PROMPT_GIT_REPO=\
"""
If you use git, I can commit the files created or
modified during setup, you will be prompted for confirmation
when that is the case.

The author name and mail address will be 'hotdoc' and
'hotdoc@hotdoc.net'.

Note: here as everywhere else, you can answer None to
skip the question.
"""

def validate_git_repo(wizard, path):
    try:
        git_interface = GitInterface(path)
        return True
    except Exception as e:
        print "This does not look like a git repo : %s" % path
        return False

PROMPT_ROOT_INDEX=\
"""
I can now create a root index to tie all your sub indexes
together, do you wish me to do that [y,n]? """

class HotdocWizard(QuickStartWizard):
    def __init__(self, *args, **kwargs):
        conf_path = kwargs.pop('conf_path', '')
        QuickStartWizard.__init__(self, *args, **kwargs)
        if self.parent == self:
            self.comments = {}
            self.symbols = {}
            self.tc = TerminalController()
            if HAVE_GIT_INTERFACE:
                self.git_interface = GitInterface()
            else:
                self.git_interface = None
            self.config = {}
            self.conf_path = conf_path
        else:
            self.comments = self.parent.comments
            self.symbols = self.parent.symbols
            self.tc = self.parent.tc
            self.git_interface = self.parent.git_interface
            self.conf_path = self.parent.conf_path

        self.tag_validators = {}

    def add_comment(self, comment):
        self.comments[comment.name] = comment

    def get_comment(self, name):
        return self.comments.get(name)

    def get_or_create_symbol(self, type_, **kwargs):
        unique_name = kwargs.get('unique_name')
        if not unique_name:
            unique_name = kwargs.get('display_name')
            kwargs['unique_name'] = unique_name

        filename = kwargs.get('filename')
        if filename:
            kwargs['filename'] = os.path.abspath(filename)

        symbol = type_()

        for key, value in kwargs.items():
            setattr(symbol, key, value)

        self.symbols[unique_name] = symbol

        return symbol

    def resolve_config_path(self, path):
        if path is None:
            return path

        res = os.path.join(self.conf_path, path)
        return os.path.abspath(res)

    def before_prompt(self):
        self.__clear_screen()

    def get_index_path(self):
        return None

    def get_index_name(self):
        return None

    def __create_root_index(self):
        contents = 'Welcome to our documentation!\n'

        for obj in self._qs_objects:
            if isinstance(obj, HotdocWizard):
                index_path = obj.get_index_path()
                index_name = obj.get_index_name()
                if index_path:
                    contents += '\n#### [%s](%s)\n' % (index_name, index_path)

        path = self.prompt_key('index_path',
                prompt='Path to save the created index in',
                store=False, validate_function=QuickStartWizard.validate_folder)

        path = os.path.join(path, 'index.markdown')

        with open (path, 'w') as f:
            f.write(contents)

        self.config['index'] = path

    def before_quick_start(self, obj):
        if type(obj) == QuickStartArgument and obj.argument.dest == 'index':
            if self.config.get('index'):
                return

            self.before_prompt() 
            if (self.ask_confirmation(PROMPT_ROOT_INDEX)):
                self.__create_root_index()
                raise Skip

    def main_prompt(self):
        prompt = self.tc.BOLD + "\nHotdoc started without arguments, starting setup\n" + self.tc.NORMAL
        prompt += self.tc.CYAN + QUICKSTART_HELP + self.tc.NORMAL

        prompt += '\nPress Enter to start setup '
        if not self.wait_for_continue(prompt):
            return False

        if not HAVE_GIT_INTERFACE:
            return True

        try:
            repo_path = self.prompt_key('git_repo', prompt=">>> Path to the git repo ? ",
                    title="the path to the root of the git repository",
                    extra_prompt=PROMPT_GIT_REPO,
                    validate_function=validate_git_repo,
                    finalize_function=HotdocWizard.finalize_path)
            self.git_interface.set_repo_path(self.resolve_config_path(repo_path))
        except Skip:
            pass

        return True

    @staticmethod
    def finalize_path(wizard, path):
        if not path:
            return path

        return os.path.relpath(path, wizard.conf_path)

    @staticmethod
    def finalize_paths(wizard, paths):
        if not paths:
            return paths

        res = []

        for path in paths:
            res.append(os.path.relpath(path, wizard.conf_path))

        return res

    def __clear_screen(self):
        sys.stdout.write(self.tc.CLEAR_SCREEN)
        sys.stdout.write(self.tc.RED + self.tc.BOLD + HOTDOC_ASCII +
                self.tc.NORMAL)

    def _add_argument_override(self, group, *args, **kwargs):
        set_default = kwargs.get('default')
        kwargs['default'] = argparse.SUPPRESS
        res = QuickStartWizard._add_argument_override(self, group, *args, **kwargs)

        if set_default:
            self.config[res.dest] = set_default

        return res

SUBCOMMAND_DESCRIPTION="""
Valid subcommands.

Exactly one subcommand is required.
Run hotdoc {subcommand} -h for more info
"""

class DocTool(object):
    def __init__(self):
        self.session = None
        self.output = None
        self.index_file = None
        self.doc_parser = None
        self.extension_classes = {}
        self.extensions = {}
        self.__comments = {}
        self.__symbols = {}
        self.tag_validators = {}
        self.raw_comment_parser = GtkDocRawCommentParser(self)
        self.link_resolver = LinkResolver(self)
        self.incremental = False
        self.comment_updated_signal = Signal()
        self.symbol_updated_signal = Signal()
        self.site_navigation = None

    def get_symbol(self, name, prefer_class=False):
        sym = self.__symbols.get(name)
        if not sym:
            sym = self.session.query(Symbol).filter(Symbol.unique_name ==
                    name).first()

        if sym:
            # Faster look up next time around
            self.__symbols[name] = sym
            sym.resolve_links(self.link_resolver)
        return sym

    def __update_symbol_comment(self, comment):
        self.session.query(Symbol).filter(Symbol.unique_name ==
                comment.name).update({'comment': comment})
        esym = self.__symbols.get(comment.name)
        if esym:
            esym.comment = comment
        self.comment_updated_signal(comment)

    def register_tag_validator(self, validator):
        self.tag_validators[validator.name] = validator

    def format_symbol(self, symbol_name):
        # FIXME this will be API, raise meaningful errors
        pages = self.doc_tree.symbol_maps.get(symbol_name)
        if not pages:
            return None

        page = pages.values()[0]

        if page.extension_name is None:
            self.formatter = HtmlFormatter(self, [])
        else:
            self.formatter = self.get_formatter(page.extension_name)

        sym = self.get_symbol(symbol_name)
        if not sym:
            return None

        self.update_doc_parser(page.extension_name)

        sym.update_children_comments()
        old_server = self.formatter.editing_server
        self.formatter.editing_server = None
        self.formatter.format_symbol(sym)
        self.formatter.editing_server = old_server

        return sym.detailed_description

    def patch_page(self, symbol, raw_comment):
        pages = self.doc_tree.symbol_maps.get(symbol.unique_name)
        if not pages:
            return False

        old_comment = symbol.comment
        new_comment = self.raw_comment_parser.parse_comment(raw_comment,
                old_comment.filename,
                old_comment.lineno,
                old_comment.lineno + raw_comment.count('\n'))

        if new_comment is None:
            return False

        if new_comment.name != symbol.unique_name:
            return False

        pages = pages.values()
        symbol.comment = new_comment
        for page in pages:
            formatter = self.get_formatter(page.extension_name)
            formatter.patch_page(page, symbol)

        return True

    def get_or_create_symbol(self, type_, **kwargs):
        unique_name = kwargs.get('unique_name')
        if not unique_name:
            unique_name = kwargs.get('display_name')
            kwargs['unique_name'] = unique_name

        filename = kwargs.get('filename')
        if filename:
            kwargs['filename'] = os.path.abspath(filename)

        symbol = self.session.query(type_).filter(type_.unique_name == unique_name).first()

        if not symbol:
            symbol = type_()
            self.session.add(symbol)

        for key, value in kwargs.items():
            setattr(symbol, key, value)

        if not symbol.comment:
            symbol.comment = Comment(symbol.unique_name)
            self.add_comment(symbol.comment)

        symbol.resolve_links(self.link_resolver)

        if self.incremental:
            self.symbol_updated_signal(symbol)

        self.__symbols[unique_name] = symbol

        return symbol

    def __setup_database(self):
        if self.session is not None:
            return

        db_path = os.path.join(self.get_private_folder(), 'hotdoc.db')
        self.engine = create_engine('sqlite:///%s' % db_path)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()
        self.session.autoflush = False
        Base.metadata.create_all(self.engine)

    def __create_change_tracker(self):
        try:
            self.change_tracker = pickle.load(open(os.path.join(self.get_private_folder(),
                'change_tracker.p'), 'rb'))
            self.incremental = True
        except IOError:
            self.change_tracker = ChangeTracker()

    def get_formatter(self, extension_name):
        ext = self.extensions.get(extension_name)
        if ext:
            return ext.get_formatter(self.output_format)
        return None

    def update_doc_parser(self, extension_name):
        ext = self.extensions.get(extension_name)
        self.doc_parser = None
        if ext:
            self.doc_parser = ext.get_doc_parser()

    def setup(self, args):
        from datetime import datetime

        n = datetime.now()
        self.__setup(args)

        print "core setup done in %s" % str(datetime.now() - n)

        for extension in self.extensions.values():
            n = datetime.now()
            self.change_tracker.mark_extension_stale_sources(extension)
            extension.setup ()
            self.change_tracker.update_extension_sources_mtimes(extension)
            self.session.flush()
            print "extension %s done in %s" % (extension.EXTENSION_NAME,
                    str(datetime.now() - n))

        n = datetime.now()
        root = self.doc_tree.pages.get(self.index_file)
        self.doc_tree.resolve_symbols(self, root)
        print "symbol resolution done in %s" % str(datetime.now() - n)

        n = datetime.now()
        self.session.flush()
        print "Database persisting done in %s" % str(datetime.now() - n)

    def get_assets_path(self):
        return os.path.join(self.output, 'assets')

    def format (self):
        from datetime import datetime

        n = datetime.now()
        self.__setup_folder(self.output)
        self.formatter = HtmlFormatter(self, [])
        root = self.doc_tree.pages.get(self.index_file)

        if not self.site_navigation:
            self.site_navigation = self.formatter.format_site_navigation(root)

        self.formatter.format(root)
        print "formatting done in %s" % str(datetime.now() - n)

    def add_comment(self, comment):
        self.__comments[comment.name] = comment
        for validator in self.tag_validators.values():
            if validator.default and not validator.name in comment.tags:
                comment.tags[validator.name] = Tag(name=validator.name,
                        description=validator.default)
        if self.incremental:
            self.__update_symbol_comment (comment)

    def get_comment (self, name):
        comment = self.__comments.get(name)
        if not comment:
            esym = self.get_symbol(name)
            if esym:
                comment = esym.comment
        return comment

    def __setup (self, args):
        if os.name == 'nt':
            self.datadir = os.path.join(os.path.dirname(__file__), '..', 'share')
        else:
            self.datadir = "/usr/share"

        self.parser = \
                argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,)

        subparsers = self.parser.add_subparsers(title='commands', dest='cmd',
                description=SUBCOMMAND_DESCRIPTION)
        subparsers.required = False
        run_parser = subparsers.add_parser('run', help='run hotdoc')
        run_parser.add_argument('--conf-file', help='Path to the config file',
                dest='conf_file', default='hotdoc.json')

        conf_parser = subparsers.add_parser('conf', help='configure hotdoc')
        conf_parser.add_argument('--quickstart', help='run a quickstart wizard',
                action='store_true')
        conf_parser.add_argument('--conf-file', help='Path to the config file',
                dest='conf_file', default='hotdoc.json')

        # First pass to get the conf path
        # FIXME: subparsers is useless, remove that hack
        init_args = list(args)
        split_pos = 0
        for i, arg in enumerate(init_args):
            if arg in ['run', 'conf']:
                split_pos = i
                break

        init_args = init_args[split_pos:]
        init_args = self.parser.parse_known_args(init_args)
        conf_file = os.path.abspath(init_args[0].conf_file)
        conf_path = os.path.dirname(conf_file)
        wizard = HotdocWizard(self.parser, conf_path=conf_path)
        self.wizard = wizard

        extension_classes = get_all_extension_classes (sort=True)

        for subclass in extension_classes:
            subclass.add_arguments (self.parser)
            self.extension_classes[subclass.EXTENSION_NAME] = subclass

        self.parser.add_argument ("-i", "--index", action="store",
                dest="index", help="location of the index file",
                finalize_function=HotdocWizard.finalize_path)
        self.parser.add_argument ("-o", "--output", action="store",
                dest="output", help="where to output the rendered documentation")
        self.parser.add_argument ("--output-format", action="store",
                dest="output_format", help="format for the output",
                default="html")
        self.parser.add_argument ("-I", "--include-path", action="append",
                dest="include_paths", help="markdown include paths")
        self.parser.add_argument ("--html-theme", action="store",
                dest="html_theme", help="html theme to use",
                finalize_function=HotdocWizard.finalize_path)
        self.parser.add_argument ("-", action="store_true", no_prompt=True,
                help="Separator to allow finishing a list of arguments before a command",
                dest="whatever")
        self.parser.add_argument ("--editing-server", action="store",
                dest="editing_server", help="If editing-server is provided, an edit button will be added")

        args = self.parser.parse_args(args)
        self.load_config(args, conf_file, wizard)

        exit_now = False

        save_config = True
        if args.cmd == 'run':
            save_config = False
            self.parse_config(wizard.config)
        elif args.cmd == 'conf':
            exit_now = True
            if args.quickstart == True:
                if wizard.quick_start():
                    print "Setup complete, building the documentation now"
                    if wizard.wait_for_continue("Setup complete, press Enter to build the doc now "):
                        self.parse_config(wizard.config)
                        exit_now = False

        if save_config:
            with open(conf_file, 'w') as f:
                f.write(json.dumps(wizard.config, indent=4))

        if exit_now:
            sys.exit(0)

    def load_config(self, args, conf_file, wizard):
        try:
            with open(conf_file, 'r') as f:
                contents = f.read()
        except IOError:
            contents = '{}'

        try:
            config = json.loads(contents)
        except ValueError:
            config = {}

        wizard.config.update(config)
        cli = dict(vars(args))
        cli.pop('cmd', None)
        cli.pop('quickstart', None)
        cli.pop('conf_file', None)

        wizard.config.update(cli)

    def __setup_folder(self, folder):
        if os.path.exists (folder):
            if not os.path.isdir (folder):
                print "Folder %s exists but is not a directory" % folder
                raise ConfigError ()
        else:
            os.mkdir (folder)

    def __create_extensions (self, args):
        for ext_class in self.extension_classes.values():
            ext = ext_class(self, args)
            self.extensions[ext.EXTENSION_NAME] = ext

    def get_private_folder(self):
        return os.path.abspath('hotdoc-private')

    def resolve_config_path(self, path):
        return self.wizard.resolve_config_path(path)

    def parse_config (self, config):
        module_path = os.path.dirname(__file__)
        default_theme_path = os.path.join(module_path, '..', 'default_theme')
        default_theme_path = os.path.abspath(default_theme_path)

        self.output = config.get('output')
        self.output_format = config.get('output_format')
        self.include_paths = config.get('include_paths')
        self.html_theme_path = config.get('html_theme', default_theme_path)
        self.git_repo_path = self.resolve_config_path(config.get('git_repo'))
        self.editing_server = config.get('editing_server')

        if self.output_format not in ["html"]:
            raise ConfigError ("Unsupported output format : %s" %
                    self.output_format)

        self.__setup_folder('hotdoc-private')

        # FIXME: we might actually want not to be naive
        #if not config.index:
        #    nif = NaiveIndexFormatter (self.c_source_scanner.symbols)
        #    config.index = "tmp_markdown_files/tmp_index.markdown"

        self.index_file = self.resolve_config_path(config.get('index'))

        if self.index_file is None:
            raise ConfigError("'index' is required")

        prefix = os.path.dirname(self.index_file)
        self.doc_tree = DocTree(self, prefix)

        self.__create_extensions (config)

        self.doc_tree.build_tree(self.index_file)
        self.doc_tree.fill_symbol_maps()

        self.__create_change_tracker()
        self.__setup_database()

    def persist(self):
        self.session.commit()
        pickle.dump(self.change_tracker, open(os.path.join(self.get_private_folder(),
            'change_tracker.p'), 'wb'))
        #pickle.dump(self.args, open(os.path.join(self.get_private_folder(),
        #    'args.p'), 'wb'))
        self.doc_tree.persist()

    def finalize (self):
        for extension in self.extensions.values():
            extension.finalize()
        self.session.close()



# -*- coding: utf-8 -*-

import os
import shutil
import pygraphviz as pg

from xml.sax.saxutils import unescape
from collections import defaultdict

from .symbols import *
from ..utils.simple_signals import Signal

def all_subclasses(cls):
        return cls.__subclasses__() + [g for s in cls.__subclasses__()
                                       for g in all_subclasses(s)]

class Formatter(object):
    formatting_page_signal = Signal()

    def __init__ (self, doc_tool):
        self.doc_tool = doc_tool
        self._output = doc_tool.output

        self.formatting_symbol_signal = Signal()

    def _create_hierarchy_graph (self, hierarchy):
        # FIXME: handle multiple inheritance
        graph = pg.AGraph(directed=True, strict=True)

        for pair in hierarchy:
            parent_link = pair[0].get_type_link()
            child_link = pair[1].get_type_link()

            graph.add_node(parent_link.title, URL=parent_link.get_link(),
                    style="rounded", shape="box")
            graph.add_node(child_link.title, URL=child_link.get_link(),
                    style="rounded", shape="box")

            graph.add_edge (parent_link.title, child_link.title)

        return graph

    def patch_page(self, page, symbol):
        raise NotImplementedError

    def format_symbol (self, symbol):
        self.__format_symbols(symbol.get_children_symbols())

        # We only need to resolve qualified symbols now because
        # they're referring to an actual type, not referred to.
        if isinstance (symbol, QualifiedSymbol):
            symbol.resolve_links(self.doc_tool.link_resolver)

        res = self.formatting_symbol_signal(symbol)

        if False in res:
            return False

        symbol.formatted_doc = self.__format_doc (symbol.comment)
        out, standalone = self._format_symbol (symbol)
        symbol.detailed_description = out

        # FIXME: figure out whether this approach is desirable
        if standalone:
            self._write_page (symbol)

        return True

    def emit_formatting_page(self, page):
        Formatter.formatting_page_signal(self, page)

    def format (self, page):
        self.__format_page (page)
        self._copy_extra_files ()

    def __format_symbols(self, symbols):
        for symbol in symbols:
            if symbol is None:
                continue
            symbol.skip = not self.format_symbol(symbol)

    def __format_page (self, page):
        if page.is_stale:
            page.output_attrs = defaultdict(lambda: defaultdict(dict))
            self.emit_formatting_page(page)
            self.doc_tool.update_doc_parser(page.extension_name)
            self.__format_symbols(page.symbols)
            page.detailed_description = self.doc_tool.formatter._format_page (page)[0]
            self.doc_tool.formatter._write_page (page)

        for pagename in page.subpages:
            cpage = self.doc_tool.doc_tree.get_page(pagename)
            formatter = self.doc_tool.get_formatter(cpage.extension_name)

            # This is a bit funky, might be better to not have
            # other code use doc_tool.formatter, but a provided one.
            if formatter and formatter != self.doc_tool.formatter:
                self.doc_tool.formatter = formatter
                self.doc_tool.formatter.format(cpage)
            else:
                self.__format_page (cpage)
            self.doc_tool.formatter = self


    def _copy_extra_files (self):
        asset_path = self.doc_tool.get_assets_path()
        if not os.path.exists (asset_path):
            os.mkdir (asset_path)

        for f in self._get_extra_files():
            basename = os.path.basename (f)
            path = os.path.join (asset_path, basename)
            if os.path.isfile(f):
                shutil.copy (f, path)
            elif os.path.isdir(f):
                shutil.rmtree(path, ignore_errors=True)
                shutil.copytree(f, path)

    def _write_page (self, page):
        path = os.path.join (self._output, page.link.ref)
        with open (path, 'w') as f:
            out = page.detailed_description
            f.write (out.encode('utf-8'))

    def _format_doc_string (self, docstring):
        if not docstring:
            return ""

        out = ""
        docstring = unescape (docstring)
        rendered_text = self.doc_tool.doc_parser.translate (docstring, 'html')
        return rendered_text

    def __format_doc (self, comment):
        out = ""
        if comment and comment.description:
            out = self.doc_tool.doc_parser.translate_comment (comment, 'html')
        return out

    def _get_extra_files (self):
        return []



# Pitivi video editor
#
#       pitivi/utils/loggable.py
#
# Copyright (c) 2009, Alessandro Decina <alessandro.decina@collabora.co.uk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
# Boston, MA 02110-1301, USA.

import sys, re

class TerminalController:
    # Cursor movement:
    BOL = ''             # : Move the cursor to the beginning of the line
    UP = ''              # : Move the cursor up one line
    DOWN = ''            # : Move the cursor down one line
    LEFT = ''            # : Move the cursor left one char
    RIGHT = ''           # : Move the cursor right one char

    # Deletion:
    CLEAR_SCREEN = ''    # : Clear the screen and move to home position
    CLEAR_EOL = ''       # : Clear to the end of the line.
    CLEAR_BOL = ''       # : Clear to the beginning of the line.
    CLEAR_EOS = ''       # : Clear to the end of the screen

    # Output modes:
    BOLD = ''            # : Turn on bold mode
    BLINK = ''           # : Turn on blink mode
    DIM = ''             # : Turn on half-bright mode
    REVERSE = ''         # : Turn on reverse-video mode
    NORMAL = ''          # : Turn off all modes

    # Cursor display:
    HIDE_CURSOR = ''     # : Make the cursor invisible
    SHOW_CURSOR = ''     # : Make the cursor visible

    # Terminal size:
    COLS = None          # : Width of the terminal (None for unknown)
    LINES = None         # : Height of the terminal (None for unknown)

    # Foreground colors:
    BLACK = BLUE = GREEN = CYAN = RED = MAGENTA = YELLOW = WHITE = ''

    # Background colors:
    BG_BLACK = BG_BLUE = BG_GREEN = BG_CYAN = ''
    BG_RED = BG_MAGENTA = BG_YELLOW = BG_WHITE = ''

    _STRING_CAPABILITIES = """
    BOL=cr UP=cuu1 DOWN=cud1 LEFT=cub1 RIGHT=cuf1
    CLEAR_SCREEN=clear CLEAR_EOL=el CLEAR_BOL=el1 CLEAR_EOS=ed BOLD=bold
    BLINK=blink DIM=dim REVERSE=rev UNDERLINE=smul NORMAL=sgr0
    HIDE_CURSOR=cinvis SHOW_CURSOR=cnorm""".split()
    _COLORS = """BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE""".split()
    _ANSICOLORS = "BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE".split()

    def __init__(self, term_stream=sys.stdout):
        # Curses isn't available on all platforms
        try:
            import curses
        except ImportError:
            return

        # If the stream isn't a tty, then assume it has no capabilities.
        if not term_stream.isatty():
            return

        # Check the terminal type.  If we fail, then assume that the
        # terminal has no capabilities.
        try:
            curses.setupterm()
        except:
            return

        # Look up numeric capabilities.
        self.COLS = curses.tigetnum('cols')
        self.LINES = curses.tigetnum('lines')

        # Look up string capabilities.
        for capability in self._STRING_CAPABILITIES:
            (attrib, cap_name) = capability.split('=')
            setattr(self, attrib, self._tigetstr(cap_name) or b'')

        # Colors
        set_fg = self._tigetstr('setf')
        if set_fg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, color, curses.tparm(set_fg, i) or b'')
        set_fg_ansi = self._tigetstr('setaf')
        if set_fg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))),
                                self._ANSICOLORS):
                setattr(self, color, curses.tparm(set_fg_ansi, i) or b'')
        set_bg = self._tigetstr('setb')
        if set_bg:
            for i, color in zip(list(range(len(self._COLORS))), self._COLORS):
                setattr(self, 'BG_' + color, curses.tparm(set_bg, i) or b'')
        set_bg_ansi = self._tigetstr('setab')
        if set_bg_ansi:
            for i, color in zip(list(range(len(self._ANSICOLORS))),
                                self._ANSICOLORS):
                setattr(
                    self, 'BG_' + color, curses.tparm(set_bg_ansi, i) or b'')

    def _tigetstr(self, cap_name):
        import curses
        cap = curses.tigetstr(cap_name) or b''
        return re.sub(r'\$<\d+>[/*]?', '', cap.decode()).encode()



# -*- coding: utf-8 -*-
#!/usr/bin/env python

import CommonMark
import sys
import re
import cgi
from xml.sax.saxutils import unescape

# Lifted from g-ir-doc-tool, muahaha
class DocScanner(object):
    def __init__(self):
        specs = [
            ('!alpha', r'[a-zA-Z0-9_]+'),
            ('!alpha_dash', r'[a-zA-Z0-9_-]+'),
            ('!anything', r'.*'),
            ('note', r'\n+>\s*<<note_contents:anything>>\s*\n'),
            ('new_paragraph', r'\n\n'),
            ('new_line', r'\n'),
            ('code_start_with_language',
                r'\|\[\<!\-\-\s*language\s*\=\s*\"<<language_name:alpha>>\"\s*\-\-\>'),
            ('code_start', r'\|\['),
            ('code_end', r'\]\|'),
            ('property', r'#<<type_name:alpha>>:(<<property_name:alpha_dash>>)'),
            ('signal', r'#<<type_name:alpha>>::(<<signal_name:alpha_dash>>)'),
            ('type_name', r'#(<<type_name:alpha>>)'),
            ('enum_value', r'%(<<member_name:alpha>>)'),
            ('parameter', r'@<<param_name:alpha>>'),
            ('function_call', r'<<symbol_name:alpha>>\s*\(\)'),
            ('include', r'{{\s*<<include_name:anything>>\s*}}'),
        ]
        self.specs = self.unmangle_specs(specs)
        self.regex = self.make_regex(self.specs)

    def unmangle_specs(self, specs):
        mangled = re.compile('<<([a-zA-Z_:]+)>>')
        specdict = dict((name.lstrip('!'), spec) for name, spec in specs)

        def unmangle(spec, name=None):
            def replace_func(match):
                child_spec_name = match.group(1)

                if ':' in child_spec_name:
                    pattern_name, child_spec_name = child_spec_name.split(':', 1)
                else:
                    pattern_name = None

                child_spec = specdict[child_spec_name]
                # Force all child specs of this one to be unnamed
                unmangled = unmangle(child_spec, None)
                if pattern_name and name:
                    return '(?P<%s_%s>%s)' % (name, pattern_name, unmangled)
                else:
                    return unmangled

            return mangled.sub(replace_func, spec)

        return [(name, unmangle(spec, name)) for name, spec in specs]

    def make_regex(self, specs):
        regex = '|'.join('(?P<%s>%s)' % (name, spec) for name, spec in specs
                         if not name.startswith('!'))
        return re.compile(regex)

    def get_properties(self, name, match):
        groupdict = match.groupdict()
        properties = {name: groupdict.pop(name)}
        name = name + "_"
        for group, value in groupdict.iteritems():
            if group.startswith(name):
                key = group[len(name):]
                properties[key] = value
        return properties

    def scan(self, text):
        pos = 0
        while True:
            match = self.regex.search(text, pos)
            if match is None:
                break

            start = match.start()
            if start > pos:
                yield ('other', text[pos:start], None)

            pos = match.end()
            name = match.lastgroup
            yield (name, match.group(0), self.get_properties(name, match))

        if pos < len(text):
            yield ('other', text[pos:], None)


class GtkDocParser (object):
    def __init__(self, doc_tool=None):
        self.funcs = {
            'other': self.format_other,
            'new_line': self.format_other,
            'new_paragraph': self.format_other,
            'note': self.format_other,
            'include': self.format_other,
            'property': self.format_property,
            'signal': self.format_signal,
            'type_name': self.format_type_name,
            'enum_value': self.format_enum_value,
            'parameter': self.format_parameter,
            'function_call': self.format_function_call,
            'code_start': self.format_code_start,
            'code_start_with_language': self.format_code_start_with_language,
            'code_end': self.format_code_end,
        }

        self.doc_tool = doc_tool

        self.__md_parser = CommonMark.DocParser()
        self.__md_renderer = CommonMark.HTMLRenderer()

    def format_other(self, match, props):
        return match

    def format_property (self, match, props):
        type_name = props['type_name']
        property_name = props['property_name']

        if self.doc_tool is None:
            return u'[](%s:%s)' % (type_name, property_name)

        linkname = "%s:%s" % (type_name, property_name)
        link = self.doc_tool.link_resolver.get_named_link (linkname)

        if link:
            return u"[“%s”](%s)" % (link.title, link.get_link())
        else:
            return u"the %s's “%s” property" % (type_name, property_name)

    def format_signal (self, match, props):
        type_name = props['type_name']
        signal_name = props['signal_name']
        if self.doc_tool is None:
            return u'[](%s::%s)' % (type_name, signal_name)

        linkname = "%s::%s" % (type_name, signal_name)
        link = self.doc_tool.link_resolver.get_named_link (linkname)

        if link:
            return u"[“%s”](%s)" % (link.title, link.get_link())
        else:
            return u"the %s's “%s” signal" % (type_name, signal_name)

    def format_type_name (self, match, props):
        type_name = props['type_name']

        # It is assumed that when there is a name collision
        # between a struct and a class name, the link is to be made
        # to the class

        if self.doc_tool is None:
            return u'[](%s)' % (type_name)

        class_name = '%s::%s' % (type_name, type_name)
        link = self.doc_tool.link_resolver.get_named_link (class_name)

        if link is None:
            link = self.doc_tool.link_resolver.get_named_link (type_name)

        if link:
            return "[%s](%s)" % (link.title, link.get_link())
        else:
            return match

    def format_enum_value (self, match, props):
        member_name = props['member_name']

        if self.doc_tool is None:
            return '[](%s)' % member_name

        link = self.doc_tool.link_resolver.get_named_link (member_name)

        if link:
            return "[%s](%s)" % (link.title, link.get_link ())
        else:
            return match

    def format_parameter (self, match, props):
        param_name = props['param_name']
        return '_%s_' % param_name

    def format_function_call (self, match, props):
        func_name = props['symbol_name']
        if self.doc_tool is None:
            return '[](%s)' % func_name

        link = self.doc_tool.link_resolver.get_named_link (func_name)

        if link:
            return "[%s()](%s)" % (link.title, link.get_link ())
        else:
            return match

    def format_code_start (self, match, props):
        return "\n```\n"

    def format_code_start_with_language (self, match, props):
        return "\n```%s\n" % props["language_name"] 

    def format_code_end (self, match, props):
        return "\n```\n"

    def translate(self, text, format_='markdown'):
        out = u''

        if not text:
            return out

        _scanner = DocScanner ()
        tokens = _scanner.scan (text)
        in_code = False
        for tok in tokens:
            kind, match, props = tok
            formatted = self.funcs[kind](match, props)
            if kind == "code_end":
                in_code = False
            if in_code:
                out += match
            else:
                out += formatted
            if kind in ["code_start", "format_code_start_with_language"]:
                in_code = True

        if format_ == 'markdown':
            return out
        elif format_ == 'html':
            out = cgi.escape(out)
            ast = self.__md_parser.parse (out.encode('utf-8'))
            rendered_text = self.__md_renderer.render(ast)
            return rendered_text

        raise Exception("Unrecognized format %s" % format_)

    def translate_comment (self, comment, format_='markdown'):
        text = unescape(comment.description)
        return self.translate (text, format_)

if __name__ == "__main__":
    gdp = GtkDocParser()
    with open (sys.argv[1], 'r') as f:
        contents = f.read ()
        out = gdp.translate (contents)
        print out

    sys.exit (0)



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import Mutable

Base = declarative_base()

class MutableDict(Mutable, dict):
    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)
            return Mutable.coerce(key, value)
        else:
            return value

    def __delitem(self, key):
        dict.__delitem__(self, key)
        self.changed()

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self.changed()

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, state):
        self.update(self)

class MutableList(Mutable, list):
    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            value = Mutable.coerce(key, value)

        return value

    def __setitem__(self, key, value):
        old_value = list.__getitem__(self, key)
        for obj, key in self._parents.items():
            old_value._parents.pop(obj, None)

        list.__setitem__(self, key, value)
        for obj, key in self._parents.items():
            value._parents[obj] = key

        self.changed()

    def append(self, item):
        list.append(self, item)
        self.changed()

    def extend(self, iterable):
        list.extend(self, iterable)
        self.changed()

    def insert(self, index, item):
        list.insert(self, index, item)
        self.changed()

    def remove(self, value):
        list.remove(self, value)
        self.changed()

    def reverse(self):
        list.reverse(self)
        self.changed()

    def pop(self, index=-1):
        item = list.pop(self, index)
        self.changed()
        return item

    def sort(self, cmp=None, key=None, reverse=False):
        list.sort(self, cmp, key, reverse)
        self.changed()

    def __getstate__(self):
        for item in self:
            if not isinstance (item, Mutable):
                continue
            for obj, key in self._parents.items():
                item._parents[obj] = key
        return list(self)

    def __setstate__(self, state):
        self[:] = state


class MutableObject(Mutable, object):
    @classmethod
    def coerce(cls, key, value):
        return value

    def __getstate__(self): 
        d = self.__dict__.copy()
        d.pop('_parents', None)
        return d

    def __setstate__(self, state):
        self.__dict__ = state

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        self.changed()



from hotdoc.core.alchemy_integration import MutableObject
from hotdoc.utils.simple_signals import Signal

class Link (MutableObject):
    resolving_link_signal = Signal()
    resolving_title_signal = Signal()
    def __init__(self, ref, title, id_):
        self._title = title
        self.ref = ref
        self.id_ = id_
        MutableObject.__init__(self)

    @property
    def title(self):
        resolved_title = Link.resolving_title_signal(self)
        resolved_title = [elem for elem in resolved_title if elem is not
                None]
        if resolved_title:
            return resolved_title[0]
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    def get_link (self):
        resolved_ref = Link.resolving_link_signal(self)
        resolved_ref = [elem for elem in resolved_ref if elem is not None]
        if resolved_ref:
            return resolved_ref[0]
        return self.ref

class LinkResolver(object):
    def __init__(self, doc_tool):
        self.__links = {}
        self.__external_links = {}
        self.doc_tool = doc_tool

    def get_named_link (self, name):
        if name in self.__links:
            return self.__links[name]

        sym = self.doc_tool.get_symbol(name)
        if sym:
            self.__links[name] = sym.link
            return sym.link

        if name in self.__external_links:
            self.__links[name] = self.__external_links[name]
            return self.__links[name]

        self.__links[name] = None
        return None

    def add_link (self, link):
        if not link.id_ in self.__links:
            self.__links[link.id_] = link

    def upsert_link (self, link, overwrite_ref=False, external=False):
        elink = self.__links.get (link.id_)

        if elink and external:
            return elink

        if external:
            self.__external_links[link.id_] = link
            return link

        if elink:
            if elink.ref is None or overwrite_ref and link.ref:
                elink.ref = link.ref
            if link.title is not None:
                elink.title = link.title
            return elink
        self.add_link (link)
        return link



# -*- coding: utf-8 -*-

import CommonMark
import os
import cgi
import cPickle as pickle
from xml.sax.saxutils import unescape

from ..utils.simple_signals import Signal
from ..utils.utils import OrderedSet

from .symbols import *
from .links import Link

class TypedSymbolsList (object):
    def __init__ (self, name):
        self.name = name
        self.symbols = []

class Page(object):
    def __init__(self, source_file):
        name = os.path.splitext(os.path.basename(source_file))[0]
        pagename = '%s.html' % name

        self.symbol_names = []
        self.subpages = OrderedSet({})
        self.link = Link (pagename, name, name) 
        self.title = None
        self.first_header = None
        self.first_paragraph = None
        self.short_description = None
        self.source_file = source_file
        self.output_attrs = None
        self.extension_name = None
        try:
            self.mtime = os.path.getmtime(source_file)
        except OSError:
            self.mtime = -1

        self.is_stale = True
        self.ast = None
        self.headers = {}


    def __getstate__(self):
        return {'symbol_names': self.symbol_names,
                'subpages': self.subpages,
                'link': self.link,
                'title': self.title,
                'first_header': self.first_header,
                'first_paragraph': self.first_paragraph,
                'short_description': self.short_description,
                'source_file': self.source_file,
                'output_attrs': None,
                'extension_name': self.extension_name,
                'ast': None,
                'headers': {},
                'is_stale': False, # At pickle time, we can assume non-staleness
                'mtime': self.mtime}

    def resolve_symbols (self, doc_tool):
        self.symbols = []

        self.typed_symbols = {}
        self.typed_symbols[Symbol] = TypedSymbolsList ('FIXME symbols')
        self.typed_symbols[FunctionSymbol] = TypedSymbolsList ("Functions")
        self.typed_symbols[CallbackSymbol] = TypedSymbolsList ("Callback Functions")
        self.typed_symbols[FunctionMacroSymbol] = TypedSymbolsList ("Function Macros")
        self.typed_symbols[ConstantSymbol] = TypedSymbolsList ("Constants")
        self.typed_symbols[ExportedVariableSymbol] = TypedSymbolsList ("Exported Variables")
        self.typed_symbols[StructSymbol] = TypedSymbolsList ("Data Structures")
        self.typed_symbols[EnumSymbol] = TypedSymbolsList ("Enumerations")
        self.typed_symbols[AliasSymbol] = TypedSymbolsList ("Aliases")
        self.typed_symbols[SignalSymbol] = TypedSymbolsList ("Signals")
        self.typed_symbols[PropertySymbol] = TypedSymbolsList ("Properties")
        self.typed_symbols[VFunctionSymbol] = TypedSymbolsList ("Virtual Methods")
        self.typed_symbols[ClassSymbol] = TypedSymbolsList ("Classes")

        self.formatted_contents = None
        self.formatted_doc = ''

        new_sym_names = []
        for sym_name in self.symbol_names:
            sym = doc_tool.get_symbol(sym_name)
            if sym:
                self.resolve_symbol (sym)

            new_symbols = sum(doc_tool.doc_tree.symbol_added_signal(self, sym),
                    [])
            for symbol in new_symbols:
                doc_tool.doc_tree.add_to_symbol_map(self, symbol)
                new_sym_names.append(symbol.unique_name)
                self.resolve_symbol (symbol)

        self.symbol_names.extend(new_sym_names)

    def resolve_symbol (self, symbol):
        if not '#' in symbol.link.ref:
            symbol.link.ref = '%s#%s' % (self.link.ref, symbol.link.ref)
        for l in symbol.get_extra_links():
            if not '#' in l.ref:
                l.ref = '%s#%s' % (self.link.ref, l.ref)
        tsl = self.typed_symbols[type(symbol)]
        tsl.symbols.append (symbol)
        self.symbols.append (symbol)
        if type (symbol) in [ClassSymbol, StructSymbol] and symbol.comment:
            if symbol.comment.short_description:
                self.short_description = symbol.comment.short_description
            if symbol.comment.title:
                self.title = symbol.comment.title

    def get_short_description (self):
        return self.short_description or self.first_paragraph

    def get_title (self):
        return self.title or self.first_header or self.link.title

class ParsedHeader(object):
    def __init__(self, ast_node, original_destination):
        self.ast_node = ast_node
        self.original_destination = original_destination

class PageParser(object):
    def __init__(self, doc_tool, doc_tree, prefix):
        self.__cmp = CommonMark.DocParser()
        self.__cmr = CommonMark.HTMLRenderer()
        self.well_known_names = {}
        self.doc_tree = doc_tree
        self.prefix = prefix
        self.doc_tool = doc_tool

    def register_well_known_name (self, wkn, callback):
        self.well_known_names[wkn] = callback

    def check_links(self, page, node, parent_node=None):
        if node.t == 'Link':
            if node.destination:
                path = os.path.join(self.prefix, node.destination)
                if not os.path.exists(path):
                    path = None
            else:
                path = None

            handler = self.well_known_names.get(node.destination)
            if handler:
                subpage, subfolder = handler(self.doc_tree)
                page.subpages.add (subpage)
                new_dest = os.path.splitext(os.path.basename(subpage))[0]
                if subfolder:
                    new_dest = subfolder + '/' + new_dest
                node.destination = '%s.html' % new_dest
            elif parent_node and parent_node.t == 'ATXHeader' and path:
                if not path in self.doc_tree.seen_pages:
                    page.subpages.add (path)
                    self.doc_tree.seen_pages.add (path)

                original_name = node.label[0].c
                parsed_header = ParsedHeader(parent_node.inline_content, path)
                page.headers[original_name] = parsed_header
                node.destination = '%s.html' % os.path.splitext(node.destination)[0]

        elif node.t == "ATXHeader" and not page.first_header:
            page.first_header = node.strings[0]

        elif node.t == "Paragraph" and not page.first_paragraph:
            page.first_paragraph = node.strings[0]

        for c in node.inline_content:
            self.check_links (page, c, node)
        for c in node.children:
            self.check_links (page, c, node)

    def parse_list (self, page, l):
        non_symbol_children = list(l.children)
        for c in l.children:
            for c2 in c.children:
                if c2.t == "Paragraph" and len (c2.inline_content) == 1:
                    if self.parse_para (page, c2):
                        non_symbol_children.remove(c)
        l.children = non_symbol_children

    def parse_para(self, page, paragraph):
        ic = paragraph.inline_content[0]

        if ic.t != "Link":
            return False

        if not ic.destination and ic.label:
            name = paragraph.strings[0].strip('[]() ')
            page.symbol_names.append(name)
            ic.destination = "not_an_actual_link_sorry"
            return True
        return False

    def parse(self, source_file):
        if not os.path.exists(source_file):
            return None

        with open(source_file, 'r') as f:
            contents = f.read()

        page = Page(source_file)

        ast = self.__cmp.parse(cgi.escape(contents))
        page.ast = ast

        for c in ast.children:
            if c.t == "List":
                self.parse_list(page, c)

        self.check_links(page, ast)

        return page

    def reparse(self, page):
        with open(page.source_file, 'r') as f:
            contents = f.read()

        ast = self.__cmp.parse(cgi.escape(contents))
        page.ast = ast

        page.symbol_names = []
        for c in ast.children:
            if c.t == "List":
                self.parse_list(page, c)

        self.check_links(page, ast)

    def parse_contents(self, page, contents):
        page.ast = self.__cmp.parse(cgi.escape(contents))

    def _update_links (self, node):
        if node.t == 'Link':
            if not hasattr(node, 'original_dest'):
                node.original_dest = node.destination
                node.original_label = node.label

            link = self.doc_tool.link_resolver.get_named_link(node.original_dest)
            if link and not node.original_label:
                node.label = []
                name_block = CommonMark.CommonMark.Block()
                name_block.c = link.title
                name_block.t = 'Str'
                node.label.append(name_block)
            if link and link.get_link() is not None:
                node.destination = link.get_link()

        for c in node.inline_content:
            self._update_links (c)
        for c in node.children:
            self._update_links (c)

    def render (self, page):
        self._update_links (page.ast)
        return self.__cmr.render (page.ast) 

    def rename_headers(self, page, new_names):
        for original_name, parsed_header in page.headers.items():
            ast_node = parsed_header.ast_node
            page = self.doc_tree.get_page(parsed_header.original_destination)

            title = page.get_title()

            if title is not None:
                ast_node[0].label[0].c = title
            elif original_name in new_names:
                ast_node[0].label[0].c = new_names[original_name]

            desc = page.get_short_description()
            if desc:
                del ast_node[1:]
                desc = self.doc_tool.doc_parser.translate (desc)
                docstring = unescape (desc)
                desc = u' — %s' % desc.encode ('utf-8')
                sub_ast = self.__cmp.parse (cgi.escape(desc))

                # I know, very specific naming
                for thing in sub_ast.children:
                    for other_thing in thing.inline_content:
                        ast_node.append (other_thing)

class DocTree(object):
    def __init__(self, doc_tool, prefix):
        self.seen_pages = set({})
        self.page_parser = PageParser(doc_tool, self, prefix)

        self.pages_path = os.path.join(doc_tool.get_private_folder(), 'pages.p')
        try:
            self.pages = pickle.load(open(self.pages_path, 'rb'))
        except:
            self.pages = {}
        self.prefix = prefix
        self.symbol_added_signal = Signal()
        doc_tool.comment_updated_signal.connect(self.__comment_updated)
        doc_tool.symbol_updated_signal.connect(self.__symbol_updated)
        self.symbol_maps = {}

    def get_page(self, name):
        return self.pages.get(name)

    def add_to_symbol_map(self, page, sym):
        symbol_map = self.symbol_maps.pop(sym.unique_name, {})
        symbol_map[page.source_file] = page
        self.symbol_maps[sym.unique_name] = symbol_map

    def fill_symbol_maps(self):
        for page in self.pages.values():
            for name in page.symbol_names:
                symbol_map = self.symbol_maps.pop(name, {})
                symbol_map[page.source_file] = page
                self.symbol_maps[name] = symbol_map

    def persist(self):
        pickle.dump(self.pages, open(self.pages_path, 'wb'))

    def build_tree (self, source_file, extension_name=None):
        page = None

        if source_file in self.pages:
            epage = self.pages[source_file]
            extension_name = epage.extension_name
            try:
                mtime = os.path.getmtime(source_file)
                if mtime == epage.mtime:
                    page = epage
            except OSError:
                page = epage

        if not page:
            page = self.page_parser.parse(source_file)
            page.extension_name = extension_name

        self.pages[source_file] = page

        for subpage in page.subpages:
            self.build_tree(subpage, extension_name=extension_name)

        return page

    def resolve_symbols(self, doc_tool, page):
        if page.is_stale:
            if page.mtime != -1 and not page.ast:
                self.page_parser.reparse(page)

            page.resolve_symbols(doc_tool)
        for pagename in page.subpages:
            cpage = self.pages[pagename]
            self.resolve_symbols(doc_tool, page=cpage)

    def __stale_symbol_pages (self, symbol_name):
        pages = self.symbol_maps.get(symbol_name, {})
        for page in pages.values():
            page.is_stale = True

    def __comment_updated(self, comment):
        self.__stale_symbol_pages(comment.name)

    def __symbol_updated(self, symbol):
        self.__stale_symbol_pages(symbol.unique_name)



from __future__ import print_function
import inspect
from weakref import WeakSet, WeakKeyDictionary


class Signal(object):
    def __init__(self):
        self._functions = WeakSet()
        self._methods = WeakKeyDictionary()

    def __call__(self, *args, **kargs):
        res = []
        # Call handler functions
        for func in self._functions:
            res.append (func(*args, **kargs))

        # Call handler methods
        for obj, funcs in self._methods.items():
            for func in funcs:
                res.append (func(obj, *args, **kargs))
        return res

    def connect(self, slot):
        if inspect.ismethod(slot):
            if slot.__self__ not in self._methods:
                self._methods[slot.__self__] = set()

            self._methods[slot.__self__].add(slot.__func__)

        else:
            self._functions.add(slot)

    def disconnect(self, slot):
        if inspect.ismethod(slot):
            if slot.__self__ in self._methods:
                self._methods[slot.__self__].remove(slot.__func__)
        else:
            if slot in self._functions:
                self._functions.remove(slot)

    def clear(self):
        self._functions.clear()
        self._methods.clear()



import os

from ..formatters.html.html_formatter import HtmlFormatter
from hotdoc.core.naive_index import NaiveIndexFormatter
from .gtk_doc_parser import GtkDocParser

class ExtDependency(object):
    def __init__(self, dependency_name, upstream=False):
        self.dependency_name = dependency_name
        self.upstream = upstream

class BaseExtension(object):
    def __init__(self, doc_tool, args):
        self.doc_tool = doc_tool
        self._formatters = {"html": HtmlFormatter(doc_tool, [])}
        self._doc_parser = GtkDocParser(doc_tool)
        self.stale_source_files = []
        self.created_symbols = {}

    @staticmethod
    def add_arguments (parser):
        pass

    def get_doc_parser (self):
        return self._doc_parser

    def get_formatter (self, output_format):
        return self._formatters.get (output_format)

    def setup (self):
        pass

    def finalize (self):
        pass

    def get_source_files(self):
        return []

    def set_stale_source_files(self, stale):
        self.stale_source_files = stale

    @staticmethod
    def get_dependencies():
        return []

    def get_or_create_symbol(self, *args, **kwargs):
        sym = self.doc_tool.get_or_create_symbol(*args, **kwargs)

        if sym:
            self.created_symbols[sym.unique_name] = sym

        return sym

    def create_naive_index(self):
        directory = self.EXTENSION_NAME + "_gen_markdown_files"
        index_name = self.EXTENSION_NAME + "-index.markdown"
        nif = NaiveIndexFormatter(self.created_symbols, directory=directory,
                index_name=index_name)
        return os.path.join(directory, index_name)



# -*- coding: utf-8 -*-

import os
import cgi
import re
import tempfile
import shutil

from wheezy.template.engine import Engine
from wheezy.template.ext.core import CoreExtension
from wheezy.template.ext.code import CodeExtension
from wheezy.template.loader import FileLoader

from ...core.symbols import *
from ...core.base_formatter import Formatter
from ...core.links import Link

class TocSection (object):
    def __init__(self, summaries, name):
        self.summaries = summaries
        self.name = name
        self.id_ = ''.join(name.split())

class SymbolDescriptions (object):
    def __init__(self, descriptions, name):
        self.descriptions = descriptions
        self.name = name

class HtmlFormatter (Formatter):
    def __init__(self, doc_tool, searchpath):
        Formatter.__init__(self, doc_tool)

        self._symbol_formatters = {
                FunctionSymbol: self._format_function,
                FunctionMacroSymbol: self._format_function_macro,
                CallbackSymbol: self._format_callback,
                ConstantSymbol: self._format_constant,
                ExportedVariableSymbol: self._format_constant,
                AliasSymbol: self._format_alias,
                StructSymbol: self._format_struct,
                EnumSymbol: self._format_enum,
                ParameterSymbol: self._format_parameter_symbol,
                ReturnValueSymbol : self._format_return_value_symbol,
                FieldSymbol: self._format_field_symbol,
                SignalSymbol: self._format_signal_symbol,
                VFunctionSymbol: self._format_vfunction_symbol,
                PropertySymbol: self._format_property_symbol,
                ClassSymbol: self._format_class_symbol,
                }

        self._summary_formatters = {
                FunctionSymbol: self._format_function_summary,
                FunctionMacroSymbol: self._format_function_macro_summary,
                CallbackSymbol: self._format_callback_summary,
                ConstantSymbol: self._format_constant_summary,
                ExportedVariableSymbol: self._format_exported_variable_summary,
                AliasSymbol: self._format_alias_summary,
                StructSymbol: self._format_struct_summary,
                EnumSymbol: self._format_enum_summary,
                SignalSymbol: self._format_signal_summary,
                VFunctionSymbol: self._format_vfunction_summary,
                PropertySymbol: self._format_property_summary,
                ClassSymbol: self._format_class_summary,
                }

        self._ordering = [ClassSymbol, FunctionSymbol, FunctionMacroSymbol, SignalSymbol,
                PropertySymbol, StructSymbol, VFunctionSymbol, EnumSymbol, ConstantSymbol,
                ExportedVariableSymbol, AliasSymbol, CallbackSymbol]

        module_path = os.path.dirname(__file__)

        self.__theme_path = doc_tool.html_theme_path
        self.__parse_theme(searchpath)

        searchpath.append (os.path.join(module_path, "templates"))
        self.engine = Engine(
            loader=FileLoader(searchpath, encoding='UTF-8'),
            extensions=[CoreExtension(), CodeExtension()]
        )

        self.editing_server = doc_tool.editing_server

        self.all_scripts = set()

    def __parse_theme(self, searchpath):
        if not self.__theme_path:
            return
        searchpath.insert (0, os.path.join(self.__theme_path, 'templates'))

    def _get_extension (self):
        return "html"

    def _format_link (self, link, title):
        out = ''
        if not link:
            print "Issue here plz check", title
            return title

        template = self.engine.get_template('link.html')
        out += '%s' % template.render ({'link': link,
                                         'link_title': title})
        return out

    def _format_type_tokens (self, type_tokens):
        out = ''
        link_before = False

        for tok in type_tokens:
            if isinstance (tok, Link):
                ref = tok.get_link()
                if ref:
                    out += self._format_link (ref, tok.title)
                    link_before = True
                else:
                    if link_before:
                        out += ' '
                    out += tok.title
                    link_before = False
            else:
                if link_before:
                    out += ' '
                out += tok
                link_before = False

        return out

    def _format_linked_symbol (self, symbol):
        out = ""

        if isinstance (symbol, QualifiedSymbol):
            out += self._format_type_tokens (symbol.type_tokens)

        # FIXME : ugly
        elif hasattr (symbol, "link"):
            out += self._format_link (symbol.link.get_link(), symbol.link.title)

        if type (symbol) == ParameterSymbol:
            out += ' ' + symbol.argname

        if type (symbol) == FieldSymbol and symbol.member_name:
            template = self.engine.get_template('inline_code.html')
            member_name = template.render ({'code': symbol.member_name})
            if symbol.is_function_pointer:
                out = member_name
                out += "()"
            else:
                out += ' ' + member_name

        return out

    def _format_callable_prototype (self, return_value, function_name,
            parameters, is_pointer):
        template = self.engine.get_template('callable_prototype.html')

        return template.render ({'return_value': return_value,
                                 'name': function_name,
                                 'parameters': parameters,
                                 'is_pointer': is_pointer,
                                })

    def __format_parameter_detail (self, name, detail, extra={}):
        template = self.engine.get_template('parameter_detail.html')
        return template.render ({'name': name,
                                 'detail': detail,
                                 'extra': extra
                                })

    def _format_callable_summary (self, callable_, return_value, function_name,
            is_callable, is_pointer):
        template = self.engine.get_template('callable_summary.html')

        return template.render({'symbol': callable_,
                                'return_value': return_value,
                                'function_name': function_name,
                                'is_callable': is_callable,
                                'is_pointer': is_pointer,
                               })

    def _format_function_summary (self, func):
        return self._format_callable_summary (
                func,
                self._format_linked_symbol (func.return_value),
                self._format_linked_symbol (func),
                True,
                False)

    def _format_callback_summary (self, callback):
        return self._format_callable_summary (
                callback,
                self._format_linked_symbol (callback.return_value),
                self._format_linked_symbol (callback),
                True,
                True)

    # FIXME : C-specific
    def _format_function_macro_summary (self, func):
        return self._format_callable_summary (
                func,
                "#define ",
                self._format_linked_symbol (func),
                True,
                False)

    def _format_constant_summary (self, constant):
        template = self.engine.get_template('constant_summary.html')
        constant_link = self._format_linked_symbol (constant)

        return template.render({'symbol': constant,
                                'constant': constant_link})

    def _format_exported_variable_summary (self, extern):
        template = self.engine.get_template('exported_variable_summary.html')
        type_link = self._format_linked_symbol(extern.type_qs)
        extern_link = self._format_linked_symbol (extern)

        return template.render({'symbol': extern,
                                'type_link': type_link,
                                'extern': extern_link})

    def _format_alias_summary (self, alias):
        template = self.engine.get_template('alias_summary.html')
        alias_link = self._format_linked_symbol (alias)

        return template.render({'symbol': alias,
                                'alias': alias_link})

    def _format_struct_summary (self, struct):
        template = self.engine.get_template('struct_summary.html')
        struct_link = self._format_linked_symbol (struct)
        return template.render({'symbol': struct,
                                'struct': struct_link})

    def _format_enum_summary (self, enum):
        template = self.engine.get_template('enum_summary.html')
        enum_link = self._format_linked_symbol (enum)
        return template.render({'symbol': enum,
                                'enum': enum_link})

    def _format_signal_summary (self, signal):
        return self._format_callable_summary (
                signal,
                self._format_linked_symbol (signal.return_value),
                self._format_linked_symbol (signal),
                True,
                False)

    def _format_vfunction_summary (self, vmethod):
        return self._format_callable_summary (
                vmethod,
                self._format_linked_symbol (vmethod.return_value),
                self._format_linked_symbol (vmethod),
                True,
                True)

    def _format_property_summary (self, prop):
        template = self.engine.get_template('property_summary.html')
        property_type = self._format_linked_symbol (prop.prop_type)

        prop_link = self._format_linked_symbol (prop)

        return template.render({'symbol': prop,
                                'property_type': property_type,
                                'property_link': prop_link,
                                'extra_contents': prop.extension_contents,
                               })

    def _format_class_summary (self, klass):
        if not klass.comment:
            return ''

        short_desc = self._format_doc_string (klass.comment.short_description)
        template = self.engine.get_template('class_summary.html')
        return template.render({'symbol': klass,
                                'klass': klass,
                                'short_description': short_desc})

    def _format_summary (self, toc_sections):
        template = self.engine.get_template('summary.html')

        return template.render({'toc_sections': toc_sections,
                               })

    def _format_symbols_toc_section (self, symbols_type, symbols_list):
        summary_formatter = self._summary_formatters.get(symbols_type)

        toc_section_summaries = []
        detailed_descriptions = []
        
        for element in symbols_list.symbols:
            if element.skip:
                continue
            if summary_formatter:
                summary = summary_formatter(element)
                if summary:
                    toc_section_summaries.append (summary)
            if element.detailed_description:
                detailed_descriptions.append (element.detailed_description)

        symbol_type = symbols_list.name

        toc_section = None
        if toc_section_summaries:
            toc_section = TocSection (toc_section_summaries, symbol_type)

        symbol_descriptions = None
        if detailed_descriptions:
            symbol_descriptions = SymbolDescriptions (detailed_descriptions,
                    symbol_type)

        return (toc_section, symbol_descriptions)

    def _format_editing_link(self, symbol):
        if not self.editing_server:
            return None

        template = self.engine.get_template("editing_link.html")
        return template.render({"symbol": symbol,
                                "editing_server": self.editing_server})

    def _format_struct (self, struct):
        raw_code = None
        if struct.raw_text is not None:
            raw_code = self._format_raw_code (struct.raw_text)

        members_list = self._format_members_list (struct.members, 'Fields')

        template = self.engine.get_template ("struct.html")
        out = template.render ({"symbol": struct,
                                "editing_link": self._format_editing_link(struct),
                                "struct": struct,
                                "raw_code": raw_code,
                                "members_list": members_list})
        return (out, False)

    def _format_enum (self, enum):
        for member in enum.members:
            template = self.engine.get_template ("enum_member.html")
            member.detailed_description = template.render ({
                                    'link': member.link,
                                    'detail': member.formatted_doc,
                                    'value': str (member.enum_value)})

        members_list = self._format_members_list (enum.members, 'Members')
        template = self.engine.get_template ("enum.html")
        out = template.render ({"symbol": enum,
                                "editing_link": self._format_editing_link(enum),
                                "enum": enum,
                                "members_list": members_list})
        return (out, False)

    def emit_formatting_page(self, page):
        page.output_attrs['html']['scripts'] = set()
        return Formatter.emit_formatting_page(self, page)

    def _format_page(self, page):
        page.formatted_contents = self.doc_tool.doc_tree.page_parser.render(page)

        toc_sections = []
        symbols_details = []

        for symbols_type in self._ordering:
            symbols_list = page.typed_symbols.get(symbols_type)
            if not symbols_list:
                continue

            toc_section, symbols_descriptions = \
                    self._format_symbols_toc_section (symbols_type,
                            symbols_list)

            if toc_section:
                toc_sections.append(toc_section)
            if symbols_descriptions:
                symbols_details.append (symbols_descriptions) 

        template = self.engine.get_template('page.html')

        toc = self._format_summary(toc_sections)

        scripts = page.output_attrs['html']['scripts']
        scripts_basenames = [os.path.basename(script)
                for script in scripts]

        self.all_scripts.update(scripts)

        out = template.render ({'page': page,
                                'site_navigation': self.doc_tool.site_navigation,
                                'scripts': scripts_basenames,
                                'toc': toc,
                                'assets_path': self._get_assets_path(),
                                'symbols_details': symbols_details})

        return (out, True)

    def format_site_navigation(self, root):
        template = self.engine.get_template('site_navigation.html')
        if template:
            return template.render({'root': root,
                                    'doc_tree': self.doc_tool.doc_tree,
                                   })
        return None

    def _format_prototype (self, function, is_pointer, title):
        return_value = self._format_linked_symbol (function.return_value)
        parameters = []
        for param in function.parameters:
            parameters.append (self._format_linked_symbol(param))

        return self._format_callable_prototype (return_value,
                title, parameters, is_pointer)

    def _format_raw_code (self, code):
        code = cgi.escape(code)
        template = self.engine.get_template('raw_code.html')
        return template.render ({'code': code})

    def _format_parameter_symbol (self, parameter):
        return (self.__format_parameter_detail (parameter.argname,
                parameter.formatted_doc, extra=parameter.extension_contents), False)

    def _format_field_symbol (self, field):
        field_id = self._format_linked_symbol (field) 
        return (self.__format_parameter_detail (field_id,
            field.formatted_doc), False)

    def _format_return_value_symbol(self, return_value):
        if not return_value or not return_value.formatted_doc:
            return ('', False)
        template = self.engine.get_template('return_value.html')
        return (template.render ({'return_value': return_value}), False)

    def _format_callable(self, callable_, callable_type, title,
            is_pointer=False):
        template = self.engine.get_template('callable.html')

        parameters = [p.detailed_description for p in callable_.parameters if
                p.detailed_description is not None]

        prototype = self._format_prototype (callable_, is_pointer, title)

        return_value_detail = None
        if callable_.return_value:
            return_value_detail = callable_.return_value.detailed_description

        tags = {}
        if callable_.comment:
            tags = callable_.comment.tags

        tags.pop('returns', None)

        out = template.render ({'prototype': prototype,
                                'symbol': callable_,
                                "editing_link": self._format_editing_link(callable_),
                                'return_value': return_value_detail,
                                'parameters': parameters,
                                'callable_type': callable_type,
                                'tags': tags,
                                'extra': callable_.extension_contents})

        return (out, False)

    def _format_signal_symbol (self, signal):
        title = "%s_callback" % re.sub ('-', '_', signal.link.title)
        return self._format_callable (signal, "signal", title)

    def _format_vfunction_symbol (self, vmethod):
        return self._format_callable (vmethod, "virtual method", vmethod.link.title)

    def _format_property_symbol (self, prop):
        type_link = self._format_linked_symbol (prop.prop_type)
        template = self.engine.get_template ('property_prototype.html')
        prototype = template.render ({'property_name': prop.link.title,
                                      'property_type': type_link})
        template = self.engine.get_template ('property.html')
        res = template.render ({'symbol': prop,
                                "editing_link": self._format_editing_link(prop),
                                'prototype': prototype,
                                'property': prop,
                                'extra': prop.extension_contents})
        return (res, False)

    def _format_class_symbol (self, klass):
        hierarchy = []
        children = []
        for p in klass.hierarchy:
            hierarchy.append(self._format_linked_symbol (p))
        for c in klass.children.itervalues():
            children.append(self._format_linked_symbol (c))

        if hierarchy or children:
            template = self.engine.get_template ("hierarchy.html")
            hierarchy = template.render ({'hierarchy': hierarchy,
                                        'children': children,
                                        'klass': klass})

        template = self.engine.get_template ('class.html')
        return (template.render ({'symbol': klass,
                                  "editing_link": self._format_editing_link(klass),
                                  'klass': klass,
                                  'hierarchy': hierarchy}),
                False)

    def _format_members_list(self, members, member_designation):
        template = self.engine.get_template('member_list.html')
        return template.render ({'members': members,
            'member_designation': member_designation})

    def _format_function(self, function):
        return self._format_callable (function, "method", function.link.title)

    def _format_callback (self, callback):
        return self._format_callable (callback, "callback",
                callback.link.title, is_pointer=True)

    def _format_function_macro(self, function_macro):
        template = self.engine.get_template('callable.html')
        prototype = self._format_raw_code (function_macro.original_text)

        parameters = [p.detailed_description for p in function_macro.parameters
                if p.detailed_description is not None]

        return_value_detail = None
        if function_macro.return_value:
            return_value_detail = function_macro.return_value.detailed_description

        out = template.render ({'prototype': prototype,
                                'symbol': function_macro,
                                "editing_link": self._format_editing_link(function_macro),
                                'return_value': return_value_detail,
                                'parameters': parameters,
                                'callable_type': "function macro",
                                'flags': None,
                                'tags': {},
                                'extra': function_macro.extension_contents})

        return (out, False)

    def _format_alias (self, alias):
        template = self.engine.get_template('alias.html')
        aliased_type = self._format_linked_symbol (alias.aliased_type)
        return (template.render ({'symbol': alias,
                                  "editing_link": self._format_editing_link(alias),
                                  'alias': alias,
                                  'aliased_type': aliased_type}), False)

    def _format_constant(self, constant):
        template = self.engine.get_template('constant.html')
        definition = self._format_raw_code (constant.original_text)
        out = template.render ({'symbol': constant,
                                "editing_link": self._format_editing_link(constant),
                                'definition': definition,
                                'constant': constant})
        return (out, False)

    def _format_symbol (self, symbol):
        format_function = self._symbol_formatters.get(type(symbol))
        if format_function:
            return format_function (symbol)
        return (None, False)

    def _format_object_hierarchy_symbol (self, symbol):
        dot_graph = self._create_hierarchy_graph (symbol.hierarchy)
        f = tempfile.NamedTemporaryFile(suffix='.svg', delete=False)
        dot_graph.draw(f, prog='dot', format='svg', args="-Grankdir=LR")
        f.close()
        with open (f.name, 'r') as f:
            contents = f.read()
        os.unlink(f.name)

        pagename = 'object_hierarchy.html'
        template = self.engine.get_template(pagename)
        res = template.render({'graph': contents,
                                'assets_path': self._get_assets_path(),
                                })
        return (res, False)

    def _get_assets_path(self):
        return 'assets'

    def _get_extra_files (self):
        dir_ = os.path.dirname(__file__)
        res = []

        if self.__theme_path:
            theme_files = os.listdir(self.__theme_path)
            theme_files.remove('templates')
            theme_files = [os.path.join(self.__theme_path, dir_) for dir_ in
                    theme_files]

            res.extend(theme_files)

        return res

    def _copy_extra_files(self):
        Formatter._copy_extra_files(self)

        js_path = os.path.join (self.doc_tool.output, 'assets', 'js')
        if not os.path.exists (js_path):
            os.mkdir (js_path)

        for script in self.all_scripts:
            dest = os.path.join(js_path, os.path.basename(script))
            shutil.copy(script, dest)



from setuptools import setup, find_packages
from pkg_resources import parse_version as V
from distutils.command.build import build
from distutils.core import Command
from setuptools.command.develop import develop
from setuptools.command.sdist import sdist
from setuptools.command.bdist_egg import bdist_egg

import shutil
import os
import tarfile
import subprocess

from hotdoc.utils.setup_utils import VersionList


pygit2_version = None
try:
    libgit2_version = subprocess.check_output(['pkg-config', '--modversion',
        'libgit2']).strip()
    known_libgit2_versions = VersionList([V('0.22.0'), V('0.23.0')])
    try:
        known_libgit2_version = known_libgit2_versions.find_le(V(libgit2_version))

        if known_libgit2_version == V('0.22.0'):
            pygit2_version = '0.22.1'
        elif known_libgit2_version == V('0.23.0'):
            pygit2_version = '0.23.2'
        else:
            print "WARNING: no compatible pygit version found"
            print "git integration disabled"
    except ValueError:
        print "Warning: too old libgit2 version %s" % libgit2_version
        print "git integration disabled"
except OSError as e:
    print "Error when trying to figure out the libgit2 version"
    print "pkg-config is probably not installed\n"
    print "git integration disabled"
except subprocess.CalledProcessError as e:
    print "\nError when trying to figure out the libgit2 version\n"
    print "git integration disabled"

source_dir = os.path.abspath('./')

class DownloadDefaultTemplate(Command):
    user_options = []
    description = "Download default html template"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import requests
        response = \
                requests.get('https://people.collabora.com/~meh/hotdoc_bootstrap_theme/dist.tgz')

        with open ('default_theme.tgz', 'wb') as f:
            f.write(response.content)

        tar = tarfile.open('default_theme.tgz')
        extract_path = os.path.join(source_dir, 'hotdoc')
        tar.extractall(extract_path)
        tar.close()

        extract_path = os.path.join(extract_path, 'dist')

        theme_path = os.path.join(source_dir, 'hotdoc', 'default_theme')

        shutil.rmtree(theme_path, ignore_errors=True)

        shutil.move(extract_path, theme_path)

        os.unlink('default_theme.tgz')

class CustomDevelop(develop):
    def run(self):
        self.run_command('download_default_template')
        return develop.run(self)

class CustomBuild(build):
    def run(self):
        self.run_command('download_default_template')
        return build.run(self)

class CustomSDist(sdist):
    def run(self):
        self.run_command('download_default_template')
        return sdist.run(self)

class CustomBDistEgg(bdist_egg):
    def run(self):
        self.run_command('download_default_template')
        return bdist_egg.run(self)

install_requires = [
    'cffi',
    'wheezy.template==0.1.167',
    'CommonMark==0.5.4',
    'pygraphviz==1.3.1',
    'sqlalchemy==1.0.9',
    'ipython==4.0.0',
    'toposort==1.4']

if pygit2_version is not None:
    install_requires.append('pygit2==%s' % pygit2_version)

setup(name='hotdoc',
        version='0.6.6',
        description='A documentation tool micro-framework',
        keywords='documentation',
        url='https://github.com/hotdoc/hotdoc',
        author='Mathieu Duponchelle',
        author_email='mathieu.duponchelle@opencreed.com',
        license='LGPL',
        packages=find_packages(),

        # Only fancy thing in there now, we want to download a
        # a default theme and bower is shitty.
        cmdclass = {'build': CustomBuild,
            'sdist': CustomSDist,
            'develop': CustomDevelop,
            'bdist_egg': CustomBDistEgg,
            'download_default_template': DownloadDefaultTemplate},
        scripts = ['hotdoc/hotdoc'],
        package_data = {
            'hotdoc.formatters.html': ['templates/*', 'assets/*'],
            'hotdoc': ['default_theme/templates/*',
                'default_theme/js/*',
                'default_theme/css/*',
                'default_theme/fonts/*'],
            },
        install_requires = install_requires,
        setup_requires = ['requests'],
        )



import shlex
import pkgutil, importlib, sys, os
import subprocess
import traceback
from pkg_resources import iter_entry_points
from toposort import toposort_flatten
import collections

def PkgConfig(args):
    cmd = ['pkg-config'] + shlex.split(args)
    out = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE).stdout
    line = out.readline()[:-1].split(" ")
    return filter(lambda a: a not in [' ', ''], line)

def all_subclasses(cls):
        return cls.__subclasses__() + [g for s in cls.__subclasses__()
                                       for g in all_subclasses(s)]

def get_all_extension_classes(sort):
    all_classes = {}
    deps_map = {}

    for entry_point in iter_entry_points(group='hotdoc.extensions',
            name='get_extension_classes'):
        entry_point.module_name
        try:
            activation_function = entry_point.load()
            classes = activation_function()
        except Exception as e:
            print "Failed to load %s" % entry_point.module_name, e
            continue

        for klass in classes:
            all_classes[klass.EXTENSION_NAME] = klass

    if not sort:
        return all_classes

    for klass in all_classes.values():
        deps = klass.get_dependencies()
        satisfied = True
        topodeps = set()
        for dep in deps:
            if dep.dependency_name not in all_classes:
                print "Missing dependency %s for %s" % (dep.dependency_name,
                        klass.EXTENSION_NAME)
                satisfied = False
                break
            if dep.upstream == True:
                topodeps.add(all_classes[dep.dependency_name])

        if not satisfied:
            continue

        deps_map[klass] = topodeps


    sorted_classes = toposort_flatten(deps_map)
    return sorted_classes

# Recipe from http://code.activestate.com/recipes/576694/
class OrderedSet(collections.MutableSet):

    def __init__(self, iterable=None):
        self.end = end = [] 
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:        
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)
